# XXX This test is a use case/acceptance test to help rewrite the api
# XXX and not included in the test suite.
# XXX TODO: make an example/documentation out of this test
# XXX       when the api has changed and the test passes
import pytest
from ase.build import molecule
from ase.units import Pascal, m, Bohr
from ase.data.vdw import vdw_radii
from ase.parallel import parprint
from gpaw.solvation import (
    # calculator
    SolvationGPAW,
    # cavities
    EffectivePotentialCavity,
    ADM12SmoothStepCavity,
    FG02SmoothStepCavity,
    # custom classes for the cavities
    Power12Potential,
    ElDensity,
    SSS09Density,
    # dielectric
    LinearDielectric,
    # non-el interactions
    SurfaceInteraction,
    VolumeInteraction,
    LeakedDensityInteraction,
    # surface and volume calculators
    GradientSurface,
    KB51Volume,
)
# poisson solver
from gpaw.solvation.poisson import ADM12PoissonSolver


# references for custom classes:
# KB51 = J. G. Kirkwood and F. P. Buff,
#        The Journal of Chemical Physics, vol. 19, no. 6, pp. 774--777, 1951
# FG02 = J.-L. Fattebert and F. Gygi,
#        Journal of Computational Chemistry, vol. 23, no. 6, pp. 662--666, 2002
# SSS09 = V. M. Sanchez, M. Sued, and D. A. Scherlis,
#         The Journal of Chemical Physics, vol. 131, no. 17, p. 174108, 2009
# ADM12 = O. Andreussi, I. Dabo, and N. Marzari,
#         The Journal of Chemical Physics, vol. 136, no. 6, p. 064102, 2012


@pytest.mark.skip(reason='TODO')
def test_solvation_api():
    # define some useful units (all api user units are ASE units!)
    dyn_per_cm = 1e-3 * Pascal * m
    Giga_Pascal = 1e9 * Pascal

    # GPAW params (examples)
    # ----------------------
    xc = 'PBE'
    h = 0.24
    vac = 4.0

    # general solvation params (examples)
    # -----------------------------------
    # electrostatic
    epsinf = 78.36
    # other interactions
    gamma = 72. * dyn_per_cm  # surface tension
    p = -0.1 * Giga_Pascal  # pressure
    V_leak = 1.0  # V (interaction energy E = V_leak * [charge outside cavity])
    # only for volume calculations respecting compressibility
    T = 298.15  # K  (also used for Boltzmann distribution)
    kappa_T = 4.53e-10 / Pascal

    # effective potential cavity params (examples)
    # --------------------------------------------
    u0 = 0.180  # eV
    vdw_radii2 = vdw_radii.copy()
    vdw_radii2[1] = 1.09

    def atomic_radii(atoms):
        return [vdw_radii2[n] for n in atoms.numbers]

    # density cavity params (examples)
    # --------------------------------
    # ADM12
    rhomin = 0.0001 / Bohr ** 3
    rhomax = 0.0050 / Bohr ** 3
    # FG02, SSS09
    rho0 = 0.0004 / Bohr ** 3
    beta = 1.3
    rho0_fake = 1.0 / Bohr ** 3
    beta_fake = 2.4

    atoms = molecule('H2O')
    atoms.center(vacuum=vac)

    def print_results(atoms):
        parprint('E = %.3f eV' % (atoms.get_potential_energy(), ))
        parprint('V = %.3f Ang ** 3' % (atoms.calc.get_cavity_volume(), ))
        parprint('A = %.3f Ang ** 2' % (atoms.calc.get_cavity_surface(), ))
        parprint('Forces:')
        parprint(atoms.get_forces())
        parprint('')

    # Cavity from 1 / r ** 12 effective potential
    atoms.calc = SolvationGPAW(
        xc=xc, h=h,
        cavity=EffectivePotentialCavity(
            effective_potential=Power12Potential(atomic_radii=atomic_radii,
                                                 u0=u0),
            temperature=T,
            surface_calculator=GradientSurface(),
            volume_calculator=KB51Volume(compressibility=kappa_T,
                                         temperature=T)),
        dielectric=LinearDielectric(epsinf=epsinf),
        interactions=[
            SurfaceInteraction(surface_tension=gamma),
            VolumeInteraction(pressure=p),
            LeakedDensityInteraction(voltage=V_leak)])
    print_results(atoms)

    # Cavity from electron density a la ADM12
    atoms.calc = SolvationGPAW(
        xc=xc, h=h,
        poissonsolver=ADM12PoissonSolver(),
        cavity=ADM12SmoothStepCavity(
            rhomin, rhomax, epsinf,
            density=ElDensity(),
            surface_calculator=GradientSurface(),
            volume_calculator=KB51Volume(compressibility=kappa_T,
                                         temperature=T)),
        dielectric=LinearDielectric(epsinf=epsinf),
        interactions=[
            SurfaceInteraction(surface_tension=gamma),
            VolumeInteraction(pressure=p),
            LeakedDensityInteraction(voltage=V_leak)])
    print_results(atoms)

    # Cavity from electron density a la FG02
    atoms.calc = SolvationGPAW(
        xc=xc, h=h,
        cavity=FG02SmoothStepCavity(
            rho0, beta,
            density=ElDensity(),
            surface_calculator=GradientSurface(),
            volume_calculator=KB51Volume(compressibility=kappa_T,
                                         temperature=T)),
        dielectric=LinearDielectric(epsinf=epsinf),
        interactions=[
            SurfaceInteraction(surface_tension=gamma),
            VolumeInteraction(pressure=p),
            LeakedDensityInteraction(voltage=V_leak)])
    print_results(atoms)

    # Cavity from fake electron density a la SSS09
    atoms.calc = SolvationGPAW(
        xc=xc, h=h,
        cavity=FG02SmoothStepCavity(
            rho0_fake, beta_fake,
            density=SSS09Density(atomic_radii=atomic_radii),
            surface_calculator=GradientSurface(),
            volume_calculator=KB51Volume(compressibility=kappa_T,
                                         temperature=T)),
        dielectric=LinearDielectric(epsinf=epsinf),
        interactions=[
            SurfaceInteraction(surface_tension=gamma),
            VolumeInteraction(pressure=p),
            LeakedDensityInteraction(voltage=V_leak)])
    print_results(atoms)



import numpy as np
from gpaw.directmin.etdm import random_a, get_n_occ
from ase.units import Hartree
from gpaw.mpi import world
from gpaw.io.logger import GPAWLogger
from copy import deepcopy


class Derivatives:

    def __init__(self, etdm, wfs, c_ref=None, a_vec_u=None,
                 update_c_ref=False, eps=1.0e-7, random_amat=False):
        """
        :param etdm:
        :param wfs:
        :param c_ref: reference orbitals C_ref
        :param a_vec_u: skew-Hermitian matrix A
        :param update_c_ref: if True update reference orbitals
        :param eps: finite difference displacement
        :param random_amat: if True, use random matrix A
        """

        self.eps = eps

        # initialize vectors of elements matrix A
        if a_vec_u is None:
            self.a_vec_u = {u: np.zeros_like(v) for u,
                                                    v in etdm.a_vec_u.items()}

        if random_amat:
            for kpt in wfs.kpt_u:
                u = etdm.kpointval(kpt)
                a = random_a(etdm.a_vec_u[u].shape, wfs.dtype)
                wfs.gd.comm.broadcast(a, 0)
                self.a_vec_u[u] = a

        # initialize orbitals:
        if c_ref is None:
            self.c_ref = etdm.dm_helper.reference_orbitals
        else:
            self.c_ref = c_ref

        # update ref orbitals if needed
        if update_c_ref:
            etdm.rotate_wavefunctions(wfs, self.a_vec_u, etdm.n_dim,
                                      self.c_ref)
            etdm.dm_helper.set_reference_orbitals(wfs, etdm.n_dim)
            self.c_ref = etdm.dm_helper.reference_orbitals
            self.a_vec_u = {u: np.zeros_like(v) for u,
                                                    v in etdm.a_vec_u.items()}

    def get_analytical_derivatives(self, etdm, ham, wfs, dens,
                                   what2calc='gradient'):
        """
           Calculate analytical gradient or approximation to the Hessian
           with respect to the elements of a skew-Hermitian matrix

        :param etdm:
        :param ham:
        :param wfs:
        :param dens:
        :param what2calc: calculate gradient or Hessian
        :return: analytical gradient or Hessian
        """

        assert what2calc in ['gradient', 'hessian']

        if what2calc == 'gradient':
            # calculate analytical gradient
            analytical_der = etdm.get_energy_and_gradients(self.a_vec_u,
                                                           etdm.n_dim,
                                                           ham, wfs, dens,
                                                           self.c_ref)[1]
        else:
            # Calculate analytical approximation to hessian
            analytical_der = np.hstack([etdm.get_hessian(kpt).copy()
                                        for kpt in wfs.kpt_u])
            analytical_der = construct_real_hessian(analytical_der)
            analytical_der = np.diag(analytical_der)

        return analytical_der

    def get_numerical_derivatives(self, etdm, ham, wfs, dens,
                                  what2calc='gradient'):
        """
           Calculate numerical gradient or Hessian with respect to
           the elements of a skew-Hermitian matrix using central finite
           differences

        :param etdm:
        :param ham:
        :param wfs:
        :param dens:
        :param what2calc: calculate gradient or Hessian
        :return: numerical gradient or Hessian
        """

        assert what2calc in ['gradient', 'hessian']

        # total dimensionality if matrices are real
        dim = sum([len(a) for a in self.a_vec_u.values()])
        steps = [1.0, 1.0j] if etdm.dtype == complex else [1.0]
        use_energy_or_gradient = {'gradient': 0, 'hessian': 1}

        matrix_exp = etdm.matrix_exp
        if what2calc == 'gradient':
            numerical_der = {u: np.zeros_like(v) for u,
                                                     v in self.a_vec_u.items()}
        else:
            numerical_der = np.zeros(shape=(len(steps) * dim,
                                            len(steps) * dim))
            # have to use exact gradient when Hessian is calculated
            etdm.matrix_exp = 'egdecomp'

        row = 0
        f = use_energy_or_gradient[what2calc]
        for step in steps:
            for kpt in wfs.kpt_u:
                u = etdm.kpointval(kpt)
                for i in range(len(self.a_vec_u[u])):
                    a = self.a_vec_u[u][i]

                    self.a_vec_u[u][i] = a + step * self.eps
                    fplus = etdm.get_energy_and_gradients(
                        self.a_vec_u, etdm.n_dim, ham, wfs, dens,
                        self.c_ref)[f]

                    self.a_vec_u[u][i] = a - step * self.eps
                    fminus = etdm.get_energy_and_gradients(
                        self.a_vec_u, etdm.n_dim, ham, wfs, dens,
                        self.c_ref)[f]

                    derf = apply_central_finite_difference_approx(
                        fplus, fminus, self.eps)

                    if what2calc == 'gradient':
                        numerical_der[u][i] += step * derf
                    else:
                        numerical_der[row] = construct_real_hessian(derf)

                    row += 1
                    self.a_vec_u[u][i] = a

        if what2calc == 'hessian':
            etdm.matrix_exp = matrix_exp

        return numerical_der


class Davidson(object):
    """
    Finite difference generalized Davidson partial diagonalizer to obtain a
    number of the eigenpairs with the smallest eigenvalues.

    The following array indexation convention is used:

    e: Target eigenpair
    w: Krylov subspace
    """

    def __init__(self, etdm, logfile, fd_mode=None, m=None, h=None,
                 eps=None, cap_krylov=None, gmf=False,
                 accurate_first_pdiag=True, remember_sp_order=False,
                 sp_order=None, seed=None):
        """
        :param etdm: ETDM object for which the partial eigendecomposition
                     should be performed.
        :param logfile: Name string of the Davidson log file. Use '-' for
                        stdout or None to discard.
        :param fd_mode: Finite difference mode for partial Hessian evaluation.
                        Must be one of 'central' for central FD or 'forward'
                        for forward FD. Central FD uses two e/g evaluations per
                        Davidson step and target eigenpair with an error
                        scaling as O(h^2), forward FD uses one with O(h)
                        scaling.
        :param m: Memory parameter indicating how large the Krylov space should
                  be able to become before resetting it with the Ritz vectors
                  of the previous step or terminating the calculation if
                  cap_krylov is True.
        :param h: Displacement (in radians of orbital rotation) for finite
                  difference partial Hessian calculation.
        :param eps: Convergence threshold for maximum component of the
                    residuals of the target eigenpairs.
        :param cap_krylov: If True, terminate the calculation if the Krylov
                           space contains more than m vectors.
        :param gmf: Toggle usage with generalized mode following instead of
                    stability analysis. The defaults and some actions will be
                    different.
        :param accurate_first_pdiag: Approximate the target saddle point order
                                     better by performing a more accurate first
                                     partial diagonalization step at the
                                     expense of more gradient evaluations
        :param remember_sp_order: If True the number of target eigenpairs is
                                  saved after converging the partial Hessian
                                  eigendecomposition once and recovered for all
                                  subsequent calculations. If False the number
                                  of target eigenpairs is always gathered from
                                  the diagonal Hessian approximation in ETDM.
        :param sp_order: If given use this value for the number of target
                         eigenpairs instead of gathering it from the diagonal
                         Hessian approximation in ETDM.
        :param seed: Seed for random perturbation of initial Krylov space.
        """

        self.gmf = gmf
        self.etdm = etdm
        self.fd_mode = fd_mode
        self.remember_sp_order = remember_sp_order
        self.sp_order = sp_order
        self.log_sp_order_once = True
        self.seed = seed
        self.V_w = []       # Krylov subspace
        self.C_we = []      # Preconditioner
        self.M = []         # Matrix used to get preconditioner every step
        self.W_w = []       # Hessian effect on Krylov subspace
        # Rayleigh matrix, smaller representation of the
        # diagonalization problem to solve:
        self.H_ww = []
        self.lambda_e = []  # Target eigenvalues
        self.y_e = []       # Target eigenvectors in subspace representation
        self.x_e = []       # Target eigenvectors
        self.r_e = []       # Residuals of target eigenvectors
        self.t_e = []       # Krylov space extension vectors
        self.l = None       # Number of target eigenpairs
        self.h = h
        self.m = m
        self.converged_e = False
        self.all_converged = False
        self.error_e = []
        self.n_iter = 0
        self.eigenvalues = []
        self.eigenvectors = []
        self.reset = False
        self.eps = eps
        self.grad = None
        self.cap_krylov = cap_krylov
        self.dim_u = {}
        self.dimtot = 0
        self.nocc = {}
        self.nbands = 0
        self.c_ref = []
        self.logfile = logfile
        self.logger = GPAWLogger(world)
        self.logger.fd = logfile
        self.first_run = accurate_first_pdiag
        self.lambda_w = []  # All eigenvalues
        self.y_w = []       # All eigenvectors in subspace representation
        self.x_w = []       # All eigenvectors
        self.check_inputs()

    def check_inputs(self):
        defaults = self.set_defaults()
        assert self.etdm.name == 'etdm', 'Check etdm.'
        if self.logfile is not None:
            assert type(self.logfile) == str, 'Check logfile.'
        if self.fd_mode is None:
            self.fd_mode = defaults['fd_mode']
        else:
            assert self.fd_mode in ['central', 'forward'], 'Check fd_mode.'
        if self.m is None:
            self.m = defaults['m']
        else:
            assert type(self.m) == int or np.isinf(self.m), 'Check m.'
        if self.h is None:
            self.h = defaults['h']
        else:
            assert type(self.h) == float, 'Check h.'
        if self.eps is None:
            self.eps = defaults['eps']
        else:
            assert type(self.eps) == float, 'Check eps.'
        if self.cap_krylov is None:
            self.cap_krylov = defaults['cap_krylov']
        else:
            assert type(self.cap_krylov) == bool, 'Check cap_krylov.'
        if self.remember_sp_order is None:
            self.remember_sp_order = defaults['remember_sp_order']
        else:
            assert type(self.remember_sp_order) == bool, \
                'Check remember_sp_order.'
        if self.sp_order is not None:
            assert type(self.sp_order) == int, 'Check sp_order.'

    def set_defaults(self):
        if self.gmf:
            return {'fd_mode': 'forward',
                    'm': 300,
                    'h': 1e-3,
                    'eps': 1e-2,
                    'cap_krylov': True,
                    'remember_sp_order': True}
        else:
            return {'fd_mode': 'central',
                    'm': np.inf,
                    'h': 1e-3,
                    'eps': 1e-3,
                    'cap_krylov': False,
                    'remember_sp_order': False}

    def todict(self):
        return {'name': 'Davidson',
                'logfile': self.logfile,
                'fd_mode': self.fd_mode,
                'm': self.m,
                'h': self.h,
                'eps': self.eps,
                'cap_krylov': self.cap_krylov,
                'gmf': self.gmf,
                'remember_sp_order': self.remember_sp_order,
                'sp_order': self.sp_order}

    def introduce(self):
        self.logger(
            '|-------------------------------------------------------|')
        self.logger(
            '|             Davidson partial diagonalizer             |')
        self.logger(
            '|-------------------------------------------------------|\n',
            flush=True)

    def run(self, wfs, ham, dens, use_prev=False):
        self.initialize(wfs, use_prev)
        if not self.gmf:
            self.etdm.sort_orbitals_mom(wfs)
        self.n_iter = 0
        self.c_ref = [deepcopy(wfs.kpt_u[x].C_nM)
                      for x in range(len(wfs.kpt_u))]
        if self.fd_mode == 'forward' and self.grad is None:
            self.obtain_grad_at_c_ref(wfs, ham, dens)
        while not self.all_converged:
            self.iterate(wfs, ham, dens)
        if self.remember_sp_order:
            if self.sp_order is None:
                self.determine_sp_order_from_lambda()
                self.logger(
                    'Saved target saddle point order as '
                    + str(self.sp_order) + ' for future partial '
                    'diagonalizations.', flush=True)
            elif self.log_sp_order_once:
                self.log_sp_order_once = False
                self.logger(
                    'Using target saddle point order of '
                    + str(self.sp_order) + '.', flush=True)
        if self.gmf:
            self.obtain_x_in_krylov_subspace()
        for k, kpt in enumerate(wfs.kpt_u):
            kpt.C_nM = deepcopy(self.c_ref[k])
        if not self.gmf:
            for kpt in wfs.kpt_u:
                self.etdm.sort_orbitals(ham, wfs, kpt)
        self.first_run = False

    def obtain_grad_at_c_ref(self, wfs, ham, dens):
        a_vec_u = {}
        n_dim = {}
        for k, kpt in enumerate(wfs.kpt_u):
            n_dim[k] = wfs.bd.nbands
            a_vec_u[k] = np.zeros_like(self.etdm.a_vec_u[k])
        self.grad = self.etdm.get_energy_and_gradients(
            a_vec_u, n_dim, ham, wfs, dens, self.c_ref)[1]

    def determine_sp_order_from_lambda(self):
        sp_order = 0
        for i in range(len(self.lambda_w)):
            if self.lambda_w[i] < 1e-8:
                sp_order += 1
            else:
                break
        self.sp_order = sp_order
        if self.sp_order == 0:
            self.sp_order = 1

    def obtain_x_in_krylov_subspace(self):
        self.x_w = []
        for i in range(len(self.lambda_w)):
            self.x_w.append(
                self.V_w[:, :len(self.lambda_w)] @ self.y_w[i].T)
        self.x_w = np.asarray(self.x_w).T

    def initialize(self, wfs, use_prev=False):
        """
        This is separate from __init__ since the initial Krylov space is
        obtained here every time a partial diagonalization is performed at
        different electronic coordinates.
        """

        dimz = 2 if self.etdm.dtype == complex else 1
        self.introduce()
        self.reset = False
        self.all_converged = False
        self.l = 0
        self.V_w = None
        self.nbands = wfs.bd.nbands
        appr_hess, appr_sp_order = self.estimate_spo_and_update_appr_hess(
            wfs, use_prev=use_prev)
        self.M = np.zeros(shape=self.dimtot * dimz)
        for i in range(self.dimtot * dimz):
            self.M[i] = np.real(appr_hess[i % self.dimtot])
        if self.sp_order is not None:
            self.l = self.sp_order
        else:
            self.l = appr_sp_order if self.gmf else appr_sp_order + 2
        if self.l == 0:
            self.l = 1
        if self.l > self.dimtot * dimz:
            self.l = self.dimtot * dimz
        self.W_w = None
        self.error_e = [np.inf for x in range(self.l)]
        self.converged_e = [False for x in range(self.l)]
        self.form_initial_krylov_subspace(
            wfs, appr_hess, dimz, use_prev=use_prev)
        text = 'Davidson will target the ' + str(self.l) + ' lowest eigenpairs'
        if self.sp_order is None:
            text += '.'
        else:
            text += ' as recovered from previous calculation.'
        self.logger(text, flush=True)

    def get_approximate_hessian_and_dim(self, wfs):
        appr_hess = []
        self.dimtot = 0
        for k, kpt in enumerate(wfs.kpt_u):
            hdia = self.etdm.get_hessian(kpt)
            self.dim_u[k] = len(hdia)
            self.dimtot += len(hdia)
            appr_hess += list(hdia.copy())
            self.nocc[k] = get_n_occ(kpt)
        return appr_hess

    def estimate_spo_and_update_appr_hess(self, wfs, use_prev=False):
        appr_sp_order = 0
        appr_hess = self.get_approximate_hessian_and_dim(wfs)
        if use_prev:
            for i in range(len(self.lambda_w)):
                if self.lambda_w[i] < -1e-4:
                    appr_sp_order += 1
                    if self.etdm.dtype == complex:
                        appr_hess[i] \
                            = self.lambda_w[i] + 1.0j * self.lambda_w[i]
                    else:
                        appr_hess[i] = self.lambda_w[i]
        else:
            for i in range(len(appr_hess)):
                if np.real(appr_hess[i]) < -1e-4:
                    appr_sp_order += 1
        return appr_hess, appr_sp_order

    def form_initial_krylov_subspace(
            self, wfs, appr_hess, dimz, use_prev=False):
        rng = np.random.default_rng(self.seed)
        wfs.timer.start('Initial Krylov subspace')
        if use_prev:
            self.randomize_krylov_subspace(rng, dimz)
        else:
            self.initialize_randomized_krylov_subspace(rng, dimz, appr_hess)
        wfs.timer.start('Modified Gram-Schmidt')
        self.V_w = mgs(self.V_w)
        wfs.timer.stop('Modified Gram-Schmidt')
        self.V_w = self.V_w.T
        wfs.timer.stop('Initial Krylov subspace')

    def randomize_krylov_subspace(self, rng, dimz, reps=1e-4):
        self.V_w = deepcopy(self.x_w.T[: self.l])
        for i in range(self.l):
            for k in range(self.dimtot):
                for l in range(dimz):
                    rand = np.zeros(shape=2)
                    if world.rank == 0:
                        rand[0] = rng.random()
                        rand[1] = 1 if rng.random() > 0.5 else -1
                    else:
                        rand[0] = 0.0
                        rand[1] = 0.0
                    world.broadcast(rand, 0)
                    self.V_w[i][l * self.dimtot + k] \
                        += rand[1] * reps * rand[0]

    def initialize_randomized_krylov_subspace(
            self, rng, dimz, appr_hess, reps=1e-4):
        do_conj = False

        # Just for F821
        v = None
        imin = None

        self.V_w = []
        for i in range(self.l):
            if do_conj:
                v[self.dimtot + imin] = -1.0
                do_conj = False
            else:
                v = np.zeros(self.dimtot * dimz)
                rdia = np.real(appr_hess).copy()
                imin = int(np.where(rdia == min(rdia))[0][0])
                rdia[imin] = np.inf
                v[imin] = 1.0
                if self.etdm.dtype == complex:
                    v[self.dimtot + imin] = 1.0
                    do_conj = True
            for l in range(self.dimtot):
                for m in range(dimz):
                    if l == imin:
                        continue
                    rand = np.zeros(shape=2)
                    if world.rank == 0:
                        rand[0] = rng.random()
                        rand[1] = 1 if rng.random() > 0.5 else -1
                    else:
                        rand[0] = 0.0
                        rand[1] = 0.0
                    world.broadcast(rand, 0)
                    v[m * self.dimtot + l] = rand[1] * reps * rand[0]
            self.V_w.append(v / np.linalg.norm(v))
        self.V_w = np.asarray(self.V_w)

    def iterate(self, wfs, ham, dens):
        self.t_e = []
        self.evaluate_W(wfs, ham, dens)
        wfs.timer.start('Rayleigh matrix formation')
        self.H_ww = self.V_w.T @ self.W_w
        wfs.timer.stop('Rayleigh matrix formation')
        self.n_iter += 1
        wfs.timer.start('Rayleigh matrix diagonalization')
        eigv, eigvec = np.linalg.eigh(self.H_ww)
        wfs.timer.stop('Rayleigh matrix diagonalization')
        eigvec = eigvec.T
        self.lambda_w = deepcopy(eigv)
        self.y_w = deepcopy(eigvec)
        self.lambda_e = eigv[: self.l]
        self.y_e = eigvec[: self.l]
        self.calculate_ritz_vectors(wfs)
        self.calculate_residuals(wfs)
        for i in range(self.l):
            self.error_e[i] = np.abs(self.r_e[i]).max()
        self.all_converged = True
        for i in range(self.l):
            self.converged_e[i] = self.error_e[i] < self.eps
            self.all_converged = self.all_converged and self.converged_e[i]
        if self.all_converged:
            self.eigenvalues = deepcopy(self.lambda_e)
            self.eigenvectors = deepcopy(self.x_e)
            self.log()
            return
        self.calculate_preconditioner(wfs)
        self.augment_krylov_subspace(wfs)
        self.log()

    def evaluate_W(self, wfs, ham, dens):
        wfs.timer.start('FD Hessian vector product')
        if self.W_w is None:
            self.W_w = []
            Vt = self.V_w.T
            for i in range(len(Vt)):
                self.W_w.append(self.get_fd_hessian(Vt[i], wfs, ham, dens))
            self.reset = False
        else:
            added = len(self.V_w[0]) - len(self.W_w[0])
            self.W_w = self.W_w.T.tolist()
            Vt = self.V_w.T
            for i in range(added):
                self.W_w.append(self.get_fd_hessian(
                    Vt[-added + i], wfs, ham, dens))
        self.W_w = np.asarray(self.W_w).T
        wfs.timer.stop('FD Hessian vector product')

    def calculate_ritz_vectors(self, wfs):
        wfs.timer.start('Ritz vector calculation')
        self.x_e = []
        for i in range(self.l):
            self.x_e.append(self.V_w @ self.y_e[i].T)
        self.x_e = np.asarray(self.x_e)
        wfs.timer.stop('Ritz vector calculation')

    def calculate_residuals(self, wfs):
        wfs.timer.start('Residual calculation')
        self.r_e = []
        for i in range(self.l):
            self.r_e.append(
                self.x_e[i] * self.lambda_e[i] - self.W_w @ self.y_e[i].T)
        self.r_e = np.asarray(self.r_e)
        wfs.timer.stop('Residual calculation')

    def calculate_preconditioner(self, wfs):
        n_dim = len(self.V_w)
        wfs.timer.start('Preconditioner calculation')
        self.C_we = np.zeros(shape=(self.l, n_dim))
        for i in range(self.l):
            self.C_we[i] = -np.abs(
                np.repeat(self.lambda_e[i], n_dim) - self.M) ** -1
            for l in range(len(self.C_we[i])):
                if self.C_we[i][l] > -0.1 * Hartree:
                    self.C_we[i][l] = -0.1 * Hartree
        wfs.timer.stop('Preconditioner calculation')

    def augment_krylov_subspace(self, wfs):
        wfs.timer.start('Krylov subspace augmentation')
        self.get_new_krylov_subspace_directions(wfs)
        self.V_w = np.asarray(self.V_w)
        if self.cap_krylov:
            if len(self.V_w) > self.m:
                self.logger(
                    'Krylov space exceeded maximum size. Partial '
                    'diagonalization is not fully converged. Current size '
                    'is ' + str(len(self.V_w) - len(self.t_e)) + '. Size at '
                    'next step would be ' + str(len(self.V_w)) + '.',
                    flush=True)
                self.all_converged = True
        wfs.timer.start('Modified Gram-Schmidt')
        self.V_w = mgs(self.V_w)
        wfs.timer.stop('Modified Gram-Schmidt')
        self.V_w = self.V_w.T
        wfs.timer.stop('Krylov subspace augmentation')

    def get_new_krylov_subspace_directions(self, wfs):
        wfs.timer.start('New directions')
        for i in range(self.l):
            if not self.converged_e[i] or (self.gmf and self.first_run):
                self.t_e.append(self.C_we[i] * self.r_e[i])
        self.t_e = np.asarray(self.t_e)
        if len(self.V_w[0]) <= self.m:
            self.V_w = self.V_w.T.tolist()
            for i in range(len(self.t_e)):
                self.V_w.append(self.t_e[i])
        elif not self.cap_krylov:
            self.reset = True
            self.V_w = deepcopy(self.x_e.tolist())
            for i in range(len(self.t_e)):
                self.V_w.append(self.t_e[i])
            self.W_w = None
        wfs.timer.stop('New directions')

    def log(self):
        self.logger('Dimensionality of Krylov space: '
                    + str(len(self.V_w[0]) - len(self.t_e)), flush=True)
        if self.reset:
            self.logger('Reset Krylov space', flush=True)
        self.logger('\nEigenvalues:\n', flush=True)
        text = ''
        for i in range(self.l):
            text += '%10d '
        indices = text % tuple(range(1, self.l + 1))
        self.logger(indices, flush=True)
        text = ''
        for i in range(self.l):
            text += '%10.6f '
        self.logger(text % tuple(self.lambda_e), flush=True)
        self.logger('\nResidual maximum components:\n', flush=True)
        self.logger(indices, flush=True)
        text = ''
        temp = list(np.round_(deepcopy(self.error_e), 6))
        for i in range(self.l):
            text += '%10s '
            temp[i] = str(temp[i])
            if self.converged_e[i]:
                temp[i] += 'c'
        self.logger(text % tuple(temp), flush=True)

    def get_fd_hessian(self, vin, wfs, ham, dens):
        """
        Get the dot product of the Hessian and a vector with a finite
        difference approximation.

        :param vin: The vector
        :param wfs:
        :param ham:
        :param dens:
        :return: Dot product vector of the Hessian and the vector.
        """

        v = self.h * vin
        c_ref = deepcopy(self.c_ref)
        gp = self.calculate_displaced_grad(wfs, ham, dens, c_ref, v)
        hessi = []
        if self.fd_mode == 'central':
            gm = self.calculate_displaced_grad(wfs, ham, dens, c_ref, -1.0 * v)
            for k in range(len(wfs.kpt_u)):
                hessi += list((gp[k] - gm[k]) * 0.5 / self.h)
        elif self.fd_mode == 'forward':
            for k in range(len(wfs.kpt_u)):
                hessi += list((gp[k] - self.grad[k]) / self.h)
        for k, kpt in enumerate(wfs.kpt_u):
            kpt.C_nM = c_ref[k]  # LCAO-specific
        dens.update(wfs)
        if self.etdm.dtype == complex:
            hessc = np.zeros(shape=(2 * self.dimtot))
            hessc[: self.dimtot] = np.real(hessi)
            hessc[self.dimtot:] = np.imag(hessi)
            return hessc
        else:
            return np.asarray(hessi)

    def calculate_displaced_grad(self, wfs, ham, dens, c_ref, v):
        a_vec_u = {}
        n_dim = {}
        start = 0
        end = 0
        for k in range(len(wfs.kpt_u)):
            a_vec_u[k] = np.zeros_like(self.etdm.a_vec_u[k])
            n_dim[k] = wfs.bd.nbands
            end += self.dim_u[k]
            a_vec_u[k] += v[start: end]
            if self.etdm.dtype == complex:
                a_vec_u[k] += 1.0j * v[self.dimtot + start: self.dimtot + end]
            start += self.dim_u[k]
        return self.etdm.get_energy_and_gradients(
            a_vec_u, n_dim, ham, wfs, dens, c_ref)[1]

    def break_instability(self, wfs, n_dim, c_ref, number,
                          initial_guess='displace', ham=None, dens=None):
        """
        Displaces orbital rotation coordinates in the direction of an
        instability. Uses a fixed displacement or performs a line search.

        :param wfs:
        :param n_dim:
        :param c_ref:
        :param number: Instability index
        :param initial_guess: How to displace. Can be one of the following:
        displace: Use a fixed displacement; line_search: Performs a
        backtracking line search.
        :param ham:
        :param dens:
        """

        assert self.converged_e, 'Davidson cannot break instabilities since' \
            + ' the partial eigendecomposition has not been converged.'
        assert len(self.lambda_e) >= number, 'Davidson cannot break' \
            + ' instability no. ' + str(number) + ' since this eigenpair was' \
            + 'not converged.'
        assert self.lambda_e[number - 1] < 0.0, 'Eigenvector no. ' \
            + str(number) + ' does not represent an instability.'

        step = self.etdm.line_search.max_step
        instability = step * self.x_e[number - 1]
        if initial_guess == 'displace':
            a_vec_u = self.displace(instability)
        elif initial_guess == 'line_search':
            assert ham is not None and dens is not None, 'Hamiltonian and' \
                                                         'density needed for' \
                                                         'line search.'
            a_vec_u = self.do_line_search(
                wfs, dens, ham, n_dim, c_ref, instability)
        self.etdm.rotate_wavefunctions(wfs, a_vec_u, n_dim, c_ref)

    def displace(self, instability):
        a_vec_u = deepcopy(self.etdm.a_vec_u)
        start = 0
        stop = 0
        for k in a_vec_u.keys():
            stop += self.dim_u[k]
            a_vec_u[k] = instability[start: stop]
            start += self.dim_u[k]
        return a_vec_u

    def do_line_search(self, wfs, dens, ham, n_dim, c_ref, instability):
        a_vec_u = {}
        p_vec_u = {}
        start = 0
        stop = 0
        for k in a_vec_u.keys():
            stop += self.dim_u[k]
            p_vec_u[k] = instability[start: stop]
            start += self.dim_u[k]
        phi, g_vec_u = self.etdm.get_energy_and_gradients(
            a_vec_u, n_dim, ham, wfs, dens, c_ref)
        der_phi = 0.0
        for k in g_vec_u:
            der_phi += g_vec_u[k].conj() @ p_vec_u[k]
        der_phi = der_phi.real
        der_phi = wfs.kd.comm.sum(der_phi)
        alpha = self.etdm.line_search.step_length_update(
            a_vec_u, p_vec_u, n_dim, ham, wfs, dens, c_ref,
            phi_0=phi, der_phi_0=der_phi, phi_old=None,
            der_phi_old=None, alpha_max=5.0, alpha_old=None,
            kpdescr=wfs.kd)[0]
        for k in a_vec_u.keys():
            a_vec_u[k] = alpha * p_vec_u[k]
        return a_vec_u

    def estimate_sp_order(self, calc, method='appr-hess', target_more=1):
        self.etdm.sort_orbitals_mom(calc.wfs)
        constraints_copy = deepcopy(self.etdm.constraints)
        self.etdm.constraints = [[] for _ in range(len(calc.wfs.kpt_u))]
        appr_hess, appr_sp_order = self.estimate_spo_and_update_appr_hess(
            calc.wfs)
        if method == 'full-hess':
            self.sp_order = appr_sp_order + target_more
            self.run(calc.wfs, calc.hamiltonian, calc.density)
            appr_hess, appr_sp_order = self.estimate_spo_and_update_appr_hess(
                calc.wfs, use_prev=True)
        for kpt in calc.wfs.kpt_u:
            self.etdm.sort_orbitals(calc.hamiltonian, calc.wfs, kpt)
        self.etdm.constraints = deepcopy(constraints_copy)
        return appr_sp_order


def mgs(vin):
    """
    Modified Gram-Schmidt orthonormalization

    :param vin: Set of vectors.
    :return: Orthonormal set of vectors.
    """

    v = deepcopy(vin)
    q = np.zeros_like(v)
    for i in range(len(v)):
        q[i] = v[i] / np.linalg.norm(v[i])
        for k in range(len(v)):
            v[k] = v[k] - np.dot(np.dot(q[i].T, v[k]), q[i])
    return q


def construct_real_hessian(hess):

    if hess.dtype == complex:
        hess_real = np.hstack((np.real(hess), np.imag(hess)))
    else:
        hess_real = hess

    return hess_real


def apply_central_finite_difference_approx(fplus, fminus, eps):

    if isinstance(fplus, dict) and isinstance(fminus, dict):
        assert (len(fplus) == len(fminus))
        derf = np.hstack([(fplus[k] - fminus[k]) * 0.5 / eps
                          for k in fplus.keys()])
    elif isinstance(fplus, float) and isinstance(fminus, float):
        derf = (fplus - fminus) * 0.5 / eps
    else:
        raise ValueError()

    return derf



import numpy as np

from ase.io.ulm import Reader
from gpaw.io import Writer

from gpaw.tddft.folding import Frequency
from gpaw.tddft.folding import FoldedFrequencies
from gpaw.lcaotddft.observer import TDDFTObserver
from gpaw.lcaotddft.utilities import read_uMM
from gpaw.lcaotddft.utilities import read_wuMM
from gpaw.lcaotddft.utilities import write_uMM
from gpaw.lcaotddft.utilities import write_wuMM


def generate_freq_w(foldedfreqs_f):
    freq_w = []
    for ff in foldedfreqs_f:
        for f in ff.frequencies:
            freq_w.append(Frequency(f, ff.folding, 'au'))
    return freq_w


class FrequencyDensityMatrixReader(object):
    def __init__(self, filename, ksl, kpt_u):
        self.ksl = ksl
        self.kpt_u = kpt_u
        self.reader = Reader(filename)
        tag = self.reader.get_tag()
        if tag != FrequencyDensityMatrix.ulmtag:
            raise RuntimeError('Unknown tag %s' % tag)
        self.version = self.reader.version

        # Read small vectors
        self.time = self.reader.time
        self.foldedfreqs_f = [FoldedFrequencies(**ff)
                              for ff in self.reader.foldedfreqs_f]
        self.freq_w = generate_freq_w(self.foldedfreqs_f)
        self.Nw = len(self.freq_w)

    def __getattr__(self, attr):
        if attr in ['rho0_uMM']:
            return read_uMM(self.kpt_u, self.ksl, self.reader, attr)
        if attr in ['FReDrho_wuMM', 'FImDrho_wuMM']:
            reim = attr[1:3]
            wlist = range(self.Nw)
            return self.read_FDrho(reim, wlist)

        try:
            return getattr(self.reader, attr)
        except KeyError:
            pass

        raise AttributeError('Attribute %s not defined in version %s' %
                             (repr(attr), repr(self.version)))

    def read_FDrho(self, reim, wlist):
        assert reim in ['Re', 'Im']
        attr = 'F%sDrho_wuMM' % reim
        return read_wuMM(self.kpt_u, self.ksl, self.reader, attr, wlist)

    def close(self):
        self.reader.close()


class FrequencyDensityMatrix(TDDFTObserver):
    version = 1
    ulmtag = 'FDM'

    def __init__(self,
                 paw,
                 dmat,
                 filename=None,
                 frequencies=None,
                 restart_filename=None,
                 interval=1):
        TDDFTObserver.__init__(self, paw, interval)
        self.has_initialized = False
        self.dmat = dmat
        self.filename = filename
        self.restart_filename = restart_filename
        self.world = paw.world
        self.ksl = paw.wfs.ksl
        self.kd = paw.wfs.kd
        self.kpt_u = paw.wfs.kpt_u
        self.log = paw.log
        if self.ksl.using_blacs:
            ksl_comm = self.ksl.block_comm
            kd_comm = self.kd.comm
            assert self.world.size == ksl_comm.size * kd_comm.size

        assert self.world.rank == self.ksl.world.rank

        if filename is not None:
            self.read(filename)
            return

        self.time = paw.time
        if isinstance(frequencies, FoldedFrequencies):
            frequencies = [frequencies]
        self.foldedfreqs_f = frequencies
        self.freq_w = generate_freq_w(self.foldedfreqs_f)
        self.Nw = np.sum([len(ff.frequencies) for ff in self.foldedfreqs_f])

    def initialize(self):
        if self.has_initialized:
            return

        if self.kd.gamma:
            self.rho0_dtype = float
        else:
            self.rho0_dtype = complex

        self.rho0_uMM = []
        for kpt in self.kpt_u:
            self.rho0_uMM.append(self.dmat.zeros(self.rho0_dtype))
        self.FReDrho_wuMM = []
        self.FImDrho_wuMM = []
        for w in range(self.Nw):
            self.FReDrho_wuMM.append([])
            self.FImDrho_wuMM.append([])
            for kpt in self.kpt_u:
                self.FReDrho_wuMM[-1].append(self.dmat.zeros(complex))
                self.FImDrho_wuMM[-1].append(self.dmat.zeros(complex))
        self.has_initialized = True

    def _update(self, paw):
        if paw.action == 'init':
            if self.time != paw.time:
                raise RuntimeError('Timestamp do not match with '
                                   'the calculator')
            self.initialize()
            if paw.niter == 0:
                rho_uMM = self.dmat.get_density_matrix(paw.niter)
                for u, kpt in enumerate(self.kpt_u):
                    rho_MM = rho_uMM[u]
                    if self.rho0_dtype == float:
                        assert np.max(np.absolute(rho_MM.imag)) == 0.0
                        rho_MM = rho_MM.real
                    self.rho0_uMM[u][:] = rho_MM
            return

        if paw.action == 'kick':
            return

        assert paw.action == 'propagate'

        time_step = paw.time - self.time
        self.time = paw.time

        # Complex exponentials with envelope
        exp_w = []
        for ff in self.foldedfreqs_f:
            exp_i = (np.exp(1.0j * ff.frequencies * self.time) *
                     ff.folding.envelope(self.time) * time_step)
            exp_w.extend(exp_i.tolist())

        rho_uMM = self.dmat.get_density_matrix((paw.niter, paw.action))
        for u, kpt in enumerate(self.kpt_u):
            Drho_MM = rho_uMM[u] - self.rho0_uMM[u]
            for w, exp in enumerate(exp_w):
                # Update Fourier transforms
                self.FReDrho_wuMM[w][u] += Drho_MM.real * exp
                self.FImDrho_wuMM[w][u] += Drho_MM.imag * exp

    def write_restart(self):
        if self.restart_filename is None:
            return
        self.write(self.restart_filename)

    def write(self, filename):
        self.log('%s: Writing to %s' % (self.__class__.__name__, filename))
        writer = Writer(filename, self.world, mode='w',
                        tag=self.__class__.ulmtag)
        writer.write(version=self.__class__.version)
        writer.write(time=self.time)
        writer.write(foldedfreqs_f=[ff.todict() for ff in self.foldedfreqs_f])
        write_uMM(self.kd, self.ksl, writer, 'rho0_uMM', self.rho0_uMM)
        wlist = range(self.Nw)
        write_wuMM(self.kd, self.ksl, writer, 'FReDrho_wuMM',
                   self.FReDrho_wuMM, wlist)
        write_wuMM(self.kd, self.ksl, writer, 'FImDrho_wuMM',
                   self.FImDrho_wuMM, wlist)
        writer.close()

    def read(self, filename):
        reader = FrequencyDensityMatrixReader(filename, self.ksl, self.kpt_u)
        self.time = reader.time
        self.foldedfreqs_f = reader.foldedfreqs_f
        self.freq_w = reader.freq_w
        self.Nw = reader.Nw
        self.rho0_uMM = reader.rho0_uMM
        self.rho0_dtype = self.rho0_uMM[0].dtype
        self.FReDrho_wuMM = reader.FReDrho_wuMM
        self.FImDrho_wuMM = reader.FImDrho_wuMM
        reader.close()
        self.has_initialized = True



from gpaw import GPAW, FermiDirac
from ase import Atoms
from gpaw.test import equal
import numpy as np


def test_generic_si(in_tmp_dir):
    a = 5.404
    bulk = Atoms(symbols='Si8',
                 scaled_positions=[(0, 0, 0),
                                   (0, 0.5, 0.5),
                                   (0.5, 0, 0.5),
                                   (0.5, 0.5, 0),
                                   (0.25, 0.25, 0.25),
                                   (0.25, 0.75, 0.75),
                                   (0.75, 0.25, 0.75),
                                   (0.75, 0.75, 0.25)],
                 pbc=True, cell=(a, a, a))
    n = 20
    calc = GPAW(gpts=(n, n, n),
                nbands=8 * 3,
                occupations=FermiDirac(width=0.01),
                kpts=(1, 1, 1))
    bulk.calc = calc
    e1 = bulk.get_potential_energy()
    eigs = calc.get_eigenvalues(kpt=0)
    calc.write('temp.gpw')

    calc = GPAW('temp.gpw').fixed_density()
    eigs2 = calc.get_eigenvalues(kpt=0)
    print('Orginal', eigs)
    print('Fixdensity', eigs2)
    print('Difference', eigs2 - eigs)

    assert np.fabs(eigs2 - eigs)[:-1].max() < 3e-5
    equal(e1, -36.767, 0.003)



import warnings

import numpy as np
from gpaw.mpi import MPIComm
from gpaw.new import zip
from gpaw.new.brillouin import IBZ, BZPoints
from gpaw.rotation import rotation
from gpaw.symmetry import Symmetry as OldSymmetry


def create_symmetries_object(atoms, ids=None, magmoms=None, parameters=None):
    ids = ids or [()] * len(atoms)
    if magmoms is None:
        pass
    elif magmoms.ndim == 1:
        ids = [id + (m,) for id, m in zip(ids, magmoms)]
    else:
        ids = [id + tuple(m) for id, m in zip(ids, magmoms)]
    symmetry = OldSymmetry(ids,
                           atoms.cell.complete(),
                           atoms.pbc,
                           **(parameters or {}))
    symmetry.analyze(atoms.get_scaled_positions())
    return Symmetries(symmetry)


def mat(rot_cc) -> str:
    """Convert 3x3 matrix to str.

    >>> mat([[-1, 0, 0], [0, 1, 0], [0, 0, 1]])
    '[[-1,  0,  0], [ 0,  1,  0], [ 0,  0,  1]]'

    """
    return '[[' + '], ['.join(', '.join(f'{r:2}'
                                        for r in rot_c)
                              for rot_c in rot_cc) + ']]'


class Symmetries:
    """Wrapper for old symmetry object.  Exposes what we need now."""
    def __init__(self, symmetry):
        self.symmetry = symmetry
        self.rotation_scc = symmetry.op_scc
        self.translation_sc = symmetry.ft_sc
        self.a_sa = symmetry.a_sa
        cell_cv = symmetry.cell_cv
        self.rotation_svv = np.einsum('vc, scd, dw -> svw',
                                      np.linalg.inv(cell_cv),
                                      self.rotation_scc,
                                      cell_cv)
        self.rotation_lsmm = [
            np.array([rotation(l, r_vv) for r_vv in self.rotation_svv])
            for l in range(4)]
        self._rotations = {}

    def __len__(self):
        return len(self.rotation_scc)

    def __str__(self):
        lines = ['symmetry:',
                 f'  number of symmetries: {len(self)}']
        if self.symmetry.symmorphic:
            lines.append('  rotations: [')
            for rot_cc in self.rotation_scc:
                lines.append(f'    {mat(rot_cc)},')
        else:
            nt = self.translation_sc.any(1).sum()
            lines.append(f'  number of symmetries with translation: {nt}')
            lines.append('  rotations and translations: [')
            for rot_cc, t_c in zip(self.rotation_scc, self.translation_sc):
                a, b, c = t_c
                lines.append(f'    [{mat(rot_cc)}, '
                             f'[{a:6.3f}, {b:6.3f}, {c:6.3f}]],')
        lines[-1] = lines[-1][:-1] + ']\n'
        return '\n'.join(lines)

    def reduce(self,
               bz: BZPoints,
               comm: MPIComm = None,
               strict: bool = True) -> IBZ:
        """Find irreducible set of k-points."""
        if not (self.symmetry.time_reversal or
                self.symmetry.point_group):
            N = len(bz)
            return IBZ(self,
                       bz,
                       ibz2bz=np.arange(N),
                       bz2ibz=np.arange(N),
                       weights=np.ones(N) / N)

        (_, weight_k, sym_k, time_reversal_k, bz2ibz_K, ibz2bz_k,
         bz2bz_Ks) = self.symmetry.reduce(bz.kpt_Kc, comm)

        if -1 in bz2bz_Ks:
            msg = 'Note: your k-points are not as symmetric as your crystal!'
            if strict:
                raise ValueError(msg)
            warnings.warn(msg)

        return IBZ(self, bz, ibz2bz_k, bz2ibz_K, weight_k)

    def check_positions(self, fracpos_ac):
        self.symmetry.check(fracpos_ac)

    def symmetrize_forces(self, F_av):
        return self.symmetry.symmetrize_forces(F_av)

    def rotations(self, l_j, xp=np):
        ells = tuple(l_j)
        rotation_sii = self._rotations.get(ells)
        if rotation_sii is None:
            ni = sum(2 * l + 1 for l in l_j)
            rotation_sii = np.zeros((len(self), ni, ni))
            i1 = 0
            for l in l_j:
                i2 = i1 + 2 * l + 1
                rotation_sii[:, i1:i2, i1:i2] = self.rotation_lsmm[l]
                i1 = i2
            rotation_sii = xp.asarray(rotation_sii)
            self._rotations[ells] = rotation_sii
        return rotation_sii



"""Module defining  ``Eigensolver`` classes."""
from functools import partial

import numpy as np

from gpaw.utilities.blas import axpy
from gpaw.eigensolvers.eigensolver import Eigensolver


class RMMDIIS(Eigensolver):
    """RMM-DIIS eigensolver

    It is expected that the trial wave functions are orthonormal
    and the integrals of projector functions and wave functions
    ``nucleus.P_uni`` are already calculated

    Solution steps are:

    * Subspace diagonalization
    * Calculation of residuals
    * Improvement of wave functions:  psi' = psi + lambda PR + lambda PR'
    * Orthonormalization"""

    def __init__(self, keep_htpsit=True, blocksize=None, niter=3, rtol=1e-16,
                 limit_lambda=False, use_rayleigh=False, trial_step=0.1):
        """Initialize RMM-DIIS eigensolver.

        Parameters:

        limit_lambda: dictionary
            determines if step length should be limited
            supported keys:
            'absolute':True/False limit the absolute value
            'upper':float upper limit for lambda
            'lower':float lower limit for lambda

        """

        Eigensolver.__init__(self, keep_htpsit, blocksize)
        self.niter = niter
        self.rtol = rtol
        self.limit_lambda = limit_lambda
        self.use_rayleigh = use_rayleigh
        if use_rayleigh:
            1 / 0
            self.blocksize = 1
        self.trial_step = trial_step
        self.first = True

    def todict(self):
        return {'name': 'rmm-diis', 'niter': self.niter}

    def initialize(self, wfs):
        if self.blocksize is None:
            if wfs.mode == 'pw':
                S = wfs.pd.comm.size
                # Use a multiple of S for maximum efficiency
                self.blocksize = int(np.ceil(10 / S)) * S
            else:
                self.blocksize = 10
        Eigensolver.initialize(self, wfs)

    def iterate_one_k_point(self, ham, wfs, kpt, weights):
        """Do a single RMM-DIIS iteration for the kpoint"""

        self.subspace_diagonalize(ham, wfs, kpt)

        psit = kpt.psit
        # psit2 = psit.new(buf=wfs.work_array)
        P = kpt.projections
        P2 = P.new()
        # dMP = P.new()
        # M_nn = wfs.work_matrix_nn
        # dS = wfs.setups.dS
        R = psit.new(buf=self.Htpsit_nG)

        self.timer.start('RMM-DIIS')
        if self.keep_htpsit:
            with self.timer('Calculate residuals'):
                self.calculate_residuals(kpt, wfs, ham, psit, P, kpt.eps_n,
                                         R, P2)

        def integrate(a_G, b_G):
            return np.real(wfs.integrate(a_G, b_G, global_integral=False))

        comm = wfs.gd.comm

        B = self.blocksize
        dR = R.new(dist=None, nbands=B)
        dpsit = dR.new()
        P = P.new(bcomm=None, nbands=B)
        P2 = P.new()
        errors_x = np.zeros(B)

        # Arrays needed for DIIS step
        if self.niter > 1:
            psit_diis_nxG = wfs.empty(B * self.niter, q=kpt.q)
            R_diis_nxG = wfs.empty(B * self.niter, q=kpt.q)

        Ht = partial(wfs.apply_pseudo_hamiltonian, kpt, ham)

        error = 0.0
        for n1 in range(0, wfs.bd.mynbands, B):
            n2 = n1 + B
            if n2 > wfs.bd.mynbands:
                n2 = wfs.bd.mynbands
                B = n2 - n1
                P = P.new(nbands=B)
                P2 = P.new()
                dR = dR.new(nbands=B, dist=None)
                dpsit = dR.new()

            n_x = np.arange(n1, n2)
            psitb = psit.view(n1, n2)

            with self.timer('Calculate residuals'):
                Rb = R.view(n1, n2)
                if not self.keep_htpsit:
                    psitb.apply(Ht, out=Rb)
                    psitb.matrix_elements(wfs.pt, out=P)
                    self.calculate_residuals(kpt, wfs, ham, psitb,
                                             P, kpt.eps_n[n_x], Rb, P2, n_x)

            errors_x[:] = 0.0
            for n in range(n1, n2):
                weight = weights[n]
                errors_x[n - n1] = weight * integrate(Rb.array[n - n1],
                                                      Rb.array[n - n1])
            comm.sum(errors_x)
            error += np.sum(errors_x)

            # Insert first vectors and residuals for DIIS step
            if self.niter > 1:
                # Save the previous vectors contiguously for each band
                # in the block
                psit_diis_nxG[:B * self.niter:self.niter] = psitb.array
                R_diis_nxG[:B * self.niter:self.niter] = Rb.array

            # Precondition the residual:
            with self.timer('precondition'):
                ekin_x = self.preconditioner.calculate_kinetic_energy(
                    psitb.array, kpt)
                self.preconditioner(Rb.array, kpt, ekin_x, out=dpsit.array)

            # Calculate the residual of dpsit_G, dR_G = (H - e S) dpsit_G:
            # self.timer.start('Apply Hamiltonian')
            dpsit.apply(Ht, out=dR)
            # self.timer.stop('Apply Hamiltonian')
            with self.timer('projections'):
                dpsit.matrix_elements(wfs.pt, out=P)

            with self.timer('Calculate residuals'):
                self.calculate_residuals(kpt, wfs, ham, dpsit,
                                         P, kpt.eps_n[n_x], dR, P2, n_x,
                                         calculate_change=True)

            # Find lam that minimizes the norm of R'_G = R_G + lam dR_G
            with self.timer('Find lambda'):
                RdR_x = np.array([integrate(dR_G, R_G)
                                  for R_G, dR_G in zip(Rb.array, dR.array)])
                dRdR_x = np.array([integrate(dR_G, dR_G) for dR_G in dR.array])
                comm.sum(RdR_x)
                comm.sum(dRdR_x)
                lam_x = -RdR_x / dRdR_x

            # Limit abs(lam) to [0.15, 1.0]
            if self.limit_lambda:
                upper = self.limit_lambda['upper']
                lower = self.limit_lambda['lower']
                if self.limit_lambda.get('absolute', False):
                    lam_x = np.where(np.abs(lam_x) < lower,
                                     lower * np.sign(lam_x), lam_x)
                    lam_x = np.where(np.abs(lam_x) > upper,
                                     upper * np.sign(lam_x), lam_x)
                else:
                    lam_x = np.where(lam_x < lower, lower, lam_x)
                    lam_x = np.where(lam_x > upper, upper, lam_x)

            # lam_x[:] = 0.1

            # New trial wavefunction and residual
            with self.timer('Update psi'):
                for lam, psit_G, dpsit_G, R_G, dR_G in zip(
                        lam_x, psitb.array,
                        dpsit.array, Rb.array,
                        dR.array):
                    axpy(lam, dpsit_G, psit_G)  # psit_G += lam * dpsit_G
                    axpy(lam, dR_G, R_G)  # R_G += lam** dR_G

            self.timer.start('DIIS step')
            # DIIS step
            for nit in range(1, self.niter):
                # Do not perform DIIS if error is small
                # if abs(error_block / B) < self.rtol:
                #     break

                # Update the subspace
                psit_diis_nxG[nit:B * self.niter:self.niter] = psitb.array
                R_diis_nxG[nit:B * self.niter:self.niter] = Rb.array

                # XXX Only integrals of nit old psits would be needed
                # self.timer.start('projections')
                # wfs.pt.integrate(psit_diis_nxG, P_diis_anxi, kpt.q)
                # self.timer.stop('projections')
                if nit > 1 or self.limit_lambda:
                    for ib in range(B):
                        istart = ib * self.niter
                        iend = istart + nit + 1

                        # Residual matrix
                        self.timer.start('Construct matrix')
                        R_nn = wfs.integrate(R_diis_nxG[istart:iend],
                                             R_diis_nxG[istart:iend],
                                             global_integral=True)

                        # Full matrix
                        A_nn = -np.ones((nit + 2, nit + 2), wfs.dtype)
                        A_nn[:nit + 1, :nit + 1] = R_nn[:]
                        A_nn[-1, -1] = 0.0
                        x_n = np.zeros(nit + 2, wfs.dtype)
                        x_n[-1] = -1.0
                        self.timer.stop('Construct matrix')
                        with self.timer('Linear solve'):
                            alpha_i = np.linalg.solve(A_nn, x_n)[:-1]

                        self.timer.start('Update trial vectors')
                        psitb.array[ib] = alpha_i[nit] * psit_diis_nxG[istart +
                                                                       nit]
                        Rb.array[ib] = alpha_i[nit] * R_diis_nxG[istart + nit]
                        for i in range(nit):
                            # axpy(alpha_i[i], psit_diis_nxG[istart + i],
                            #      psit_diis_nxG[istart + nit])
                            # axpy(alpha_i[i], R_diis_nxG[istart + i],
                            #      R_diis_nxG[istart + nit])
                            axpy(alpha_i[i], psit_diis_nxG[istart + i],
                                 psitb.array[ib])
                            axpy(alpha_i[i], R_diis_nxG[istart + i],
                                 Rb.array[ib])
                        self.timer.stop('Update trial vectors')

                if nit < self.niter - 1:
                    with self.timer('precondition'):
                        self.preconditioner(Rb.array, kpt,
                                            ekin_x, out=dpsit.array)

                    for psit_G, lam, dpsit_G in zip(psitb.array, lam_x,
                                                    dpsit.array):
                        axpy(lam, dpsit_G, psit_G)

                    # Calculate the new residuals
                    self.timer.start('Calculate residuals')
                    psitb.apply(Ht, out=Rb)
                    psitb.matrix_elements(wfs.pt, out=P)
                    self.calculate_residuals(kpt, wfs, ham, psitb,
                                             P, kpt.eps_n[n_x], Rb, P2, n_x,
                                             calculate_change=True)
                    self.timer.stop('Calculate residuals')

            self.timer.stop('DIIS step')
            # Final trial step
            with self.timer('precondition'):
                self.preconditioner(Rb.array, kpt, ekin_x, out=dpsit.array)

            self.timer.start('Update psi')
            if self.trial_step is not None:
                lam_x[:] = self.trial_step
            for lam, psit_G, dpsit_G in zip(lam_x, psitb.array, dpsit.array):
                axpy(lam, dpsit_G, psit_G)  # psit_G += lam * dpsit_G
            self.timer.stop('Update psi')

        self.timer.stop('RMM-DIIS')
        return error

    def __repr__(self):
        repr_string = 'RMM-DIIS eigensolver\n'
        repr_string += '       keep_htpsit: %s\n' % self.keep_htpsit
        repr_string += '       DIIS iterations: %d\n' % self.niter
        repr_string += '       Threshold for DIIS: %5.1e\n' % self.rtol
        repr_string += '       Limit lambda: %s\n' % self.limit_lambda
        repr_string += '       use_rayleigh: %s\n' % self.use_rayleigh
        repr_string += '       trial_step: %s' % self.trial_step
        return repr_string



"""Smearing functions and occupation number calculators."""

import warnings
from math import pi, nan, inf
from typing import List, Tuple, NamedTuple, Any, Callable, Dict, cast
import numpy as np
from scipy.special import erf
from ase.units import Ha

from gpaw.band_descriptor import BandDescriptor
from gpaw.mpi import serial_comm, broadcast_float
from gpaw.typing import Array1D, Array2D

# typehints:
MPICommunicator = Any


class ParallelLayout(NamedTuple):
    """Collection of parallel stuff."""
    bd: BandDescriptor
    kpt_comm: MPICommunicator
    domain_comm: MPICommunicator


def fermi_dirac(eig: np.ndarray,
                fermi_level: float,
                width: float) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Fermi-Dirac distribution function.

    >>> f, _, _ = fermi_dirac(0.0, 0.0, 0.1)
    >>> f
    0.5
    """
    x = (eig - fermi_level) / width
    x = np.clip(x, -100, 100)
    y = np.exp(x)
    z = y + 1.0
    f = 1.0 / z
    dfde = (f - f**2) / width
    y *= x
    y /= z
    y -= np.log(z)
    e_entropy = y * width
    return f, dfde, e_entropy


def marzari_vanderbilt(eig: np.ndarray,
                       fermi_level: float,
                       width: float) -> Tuple[np.ndarray,
                                              np.ndarray,
                                              np.ndarray]:
    """Marzari-Vanderbilt distribution (cold smearing).

    See: :doi:`10.1103/PhysRevLett.82.3296`
    """
    x = (eig - fermi_level) / width
    expterm = np.exp(-(x + (1 / np.sqrt(2)))**2)
    f = expterm / np.sqrt(2 * np.pi) + 0.5 * (1 - erf(1. / np.sqrt(2) + x))
    dfde = expterm * (2 + np.sqrt(2) * x) / np.sqrt(np.pi) / width
    s = expterm * (1 + np.sqrt(2) * x) / (2 * np.sqrt(np.pi))
    e_entropy = -s * width
    return f, dfde, e_entropy


def methfessel_paxton(eig: np.ndarray,
                      fermi_level: float,
                      width: float,
                      order: int = 0) -> Tuple[np.ndarray,
                                               np.ndarray,
                                               np.ndarray]:
    """Methfessel-Paxton distribution."""
    x = (eig - fermi_level) / width
    f = 0.5 * (1 - erf(x))
    for i in range(order):
        f += (coff_function(i + 1) *
              hermite_poly(2 * i + 1, x) * np.exp(-x**2))
    dfde = 1 / np.sqrt(pi) * np.exp(-x**2)
    for i in range(order):
        dfde += (coff_function(i + 1) *
                 hermite_poly(2 * i + 2, x) * np.exp(-x**2))
    dfde *= 1.0 / width
    e_entropy = (0.5 * coff_function(order) *
                 hermite_poly(2 * order, x) * np.exp(-x**2))
    e_entropy = -e_entropy * width
    return f, dfde, e_entropy


def coff_function(n):
    return (-1)**n / (np.product(np.arange(1, n + 1)) *
                      4**n * np.sqrt(np.pi))


def hermite_poly(n, x):
    if n == 0:
        return 1
    elif n == 1:
        return 2 * x
    else:
        return (2 * x * hermite_poly(n - 1, x) -
                2 * (n - 1) * hermite_poly(n - 2, x))


class OccupationNumberCalculator:
    """Base class for all occupation number calculators."""
    name = 'unknown'
    extrapolate_factor: float

    def __init__(self,
                 parallel_layout: ParallelLayout = None):
        """Object for calculating fermi level(s) and occupation numbers.

        If fixmagmom=True then the fixed_magmom_value attribute must be set
        and two fermi levels will be calculated.
        """
        if parallel_layout is None:
            parallel_layout = ParallelLayout(BandDescriptor(1),
                                             serial_comm,
                                             serial_comm)
        self.bd = parallel_layout.bd
        self.kpt_comm = parallel_layout.kpt_comm
        self.domain_comm = parallel_layout.domain_comm

    @property
    def parallel_layout(self) -> ParallelLayout:
        return ParallelLayout(self.bd, self.kpt_comm, self.domain_comm)

    def todict(self):
        return {'name': self.name}

    def copy(self,
             parallel_layout: ParallelLayout = None,
             bz2ibzmap: List[int] = None) -> 'OccupationNumberCalculator':
        return create_occ_calc(
            self.todict(),
            parallel_layout=parallel_layout or self.parallel_layout)

    def calculate(self,
                  nelectrons: float,
                  eigenvalues: List[List[float]],
                  weights: List[float],
                  fermi_levels_guess: List[float] = None) -> Tuple[Array2D,
                                                                   List[float],
                                                                   float]:
        """Calculate occupation numbers and fermi level(s) from eigenvalues.

        nelectrons:
            Number of electrons.
        eigenvalues: ndarray, shape=(nibzkpts, nbands)
            Eigenvalues in Hartree.
        weights: ndarray, shape=(nibzkpts,)
            Weights of k-points in IBZ (must sum to 1).
        parallel:
            Parallel distribution of eigenvalues.
        fermi_level_guesses:
            Optional guess(es) at fermi level(s).

        Returns a tuple containing:

        * occupation numbers (in the range 0 to 1)
        * fermi-level in Hartree
        * entropy as -S*T in Hartree

        >>> occ = ZeroWidth()
        >>> occ.calculate(1, [[0, 1]], [1])
        (array([[1., 0.]]), [0.5], 0.0)
        """

        eig_qn = [np.asarray(eig_n) for eig_n in eigenvalues]
        weight_q = np.asarray(weights)

        if fermi_levels_guess is None:
            fermi_levels_guess = [nan]

        f_qn = np.empty((len(weight_q), len(eig_qn[0])))

        result = np.empty(2)

        if self.domain_comm.rank == 0:
            # Let the master domain do the work and broadcast results:
            result[:] = self._calculate(
                nelectrons, eig_qn, weight_q, f_qn, fermi_levels_guess[0])

        self.domain_comm.broadcast(result, 0)
        self.domain_comm.broadcast(f_qn, 0)

        fermi_level, e_entropy = result
        return f_qn, [fermi_level], e_entropy

    def _calculate(self,
                   nelectrons: float,
                   eig_qn: List[Array1D],
                   weight_q: Array1D,
                   f_qn: Array2D,
                   fermi_level_guess: float) -> Tuple[float, float]:
        raise NotImplementedError


class FixMagneticMomentOccupationNumberCalculator(OccupationNumberCalculator):
    """Base class for all occupation number objects."""
    name = 'fixmagmom'

    def __init__(self,
                 occ: OccupationNumberCalculator,
                 magmom: float):
        """Object for calculating fermi level(s) and occupation numbers.

        If fixmagmom=True then the fixed_magmom_value attribute must be set
        and two fermi levels will be calculated.
        """
        self.occ = occ
        self.fixed_magmom_value = magmom
        self.extrapolate_factor = occ.extrapolate_factor

    def __str__(self):
        return (f'Fixed magnetic moment: {self.fixed_magmom_value:.3f}\n' +
                str(self.occ))

    def todict(self):
        dct = self.occ.todict()
        dct['fixmagmom'] = True
        return dct

    def calculate(self,
                  nelectrons: float,
                  eigenvalues: List[List[float]],
                  weights: List[float],
                  fermi_levels_guess: List[float] = None
                  ) -> Tuple[Array2D,
                             List[float],
                             float]:

        magmom = self.fixed_magmom_value

        if fermi_levels_guess is None:
            fermi_levels_guess = [nan, nan]

        f1_qn, fermi_levels1, e_entropy1 = self.occ.calculate(
            (nelectrons + magmom) / 2,
            eigenvalues[::2],
            weights[::2],
            fermi_levels_guess[:1])

        f2_qn, fermi_levels2, e_entropy2 = self.occ.calculate(
            (nelectrons - magmom) / 2,
            eigenvalues[1::2],
            weights[1::2],
            fermi_levels_guess[1:])

        f_qn = []
        for f1_n, f2_n in zip(f1_qn, f2_qn):
            f_qn += [f1_n, f2_n]

        return (np.array(f_qn),
                fermi_levels1 + fermi_levels2,
                e_entropy1 + e_entropy2)


class SmoothDistribution(OccupationNumberCalculator):
    """Base class for Fermi-Dirac and other smooth distributions."""
    def __init__(self, width: float, parallel_layout: ParallelLayout = None):
        """Smooth distribution.

        width: float
            Width of distribution in eV.
        fixmagmom: bool
            Fix spin moment calculations.  A separate Fermi level for
            spin up and down electrons is found.
        """

        self._width = width
        OccupationNumberCalculator.__init__(self, parallel_layout)

    def todict(self):
        return {'name': self.name, 'width': self._width}

    def _calculate(self,
                   nelectrons,
                   eig_qn,
                   weight_q,
                   f_qn,
                   fermi_level_guess):
        # Guess can be nan or inf:
        if not np.isfinite(fermi_level_guess) or self._width == 0.0:
            zero = ZeroWidth(self.parallel_layout)
            fermi_level_guess, _ = zero._calculate(
                nelectrons, eig_qn, weight_q, f_qn)
            if self._width == 0.0 or np.isinf(fermi_level_guess):
                return fermi_level_guess, 0.0

        x = fermi_level_guess

        data = np.empty(3)

        def func(x, data=data):
            data[:] = 0.0
            for eig_n, weight, f_n in zip(eig_qn, weight_q, f_qn):
                f_n[:], dfde_n, e_entropy_n = self.distribution(eig_n, x)
                data += [weight * x_n.sum()
                         for x_n in [f_n, dfde_n, e_entropy_n]]
            self.bd.comm.sum(data)
            self.kpt_comm.sum(data)
            f, dfde = data[:2]
            df = f - nelectrons
            return df, dfde

        fermi_level, niter = findroot(func, x)

        e_entropy = data[2]

        return fermi_level, e_entropy


class FermiDiracCalculator(SmoothDistribution):
    name = 'fermi-dirac'
    extrapolate_factor = -0.5

    def distribution(self,
                     eig_n: np.ndarray,
                     fermi_level: float) -> Tuple[np.ndarray,
                                                  np.ndarray,
                                                  np.ndarray]:
        return fermi_dirac(eig_n, fermi_level, self._width)

    def __str__(self):
        return f'Fermi-Dirac:\n  width: {self._width:.4f}  # eV\n'


class MarzariVanderbiltCalculator(SmoothDistribution):
    name = 'marzari-vanderbilt'
    # According to Nicola Marzari, one should not extrapolate M-V energies
    # https://lists.quantum-espresso.org/pipermail/users/2005-October/003170.html
    extrapolate_factor = 0.0

    def distribution(self, eig_n, fermi_level):
        return marzari_vanderbilt(eig_n, fermi_level, self._width)

    def __str__(self):
        return f'Marzari-Vanderbilt:\n  width: {self._width:.4f}  # eV\n'


class MethfesselPaxtonCalculator(SmoothDistribution):
    name = 'methfessel_paxton'

    def __init__(self, width, order=0, parallel_layout: ParallelLayout = None):
        SmoothDistribution.__init__(self, width, parallel_layout)
        self.order = order
        self.extrapolate_factor = -1.0 / (self.order + 2)

    def todict(self):
        dct = SmoothDistribution.todict(self)
        dct['order'] = self.order
        return dct

    def __str__(self):
        return ('Methfessel-Paxton:\n'
                f'  width: {self._width:.4f}  # eV\n'
                f'  order: {self.order}\n')

    def distribution(self, eig_n, fermi_level):
        return methfessel_paxton(eig_n, fermi_level, self._width, self.order)


def findroot(func: Callable[[float], Tuple[float, float]],
             x: float,
             tol: float = 1e-10) -> Tuple[float, int]:
    """Function used for locating Fermi level.

    The function should return a (value, derivative) tuple:

    >>> x, _ = findroot(lambda x: (x, 1.0), 1.0)
    >>> assert abs(x) < 1e-10
    """

    assert np.isfinite(x), x

    xmin = -np.inf
    xmax = np.inf

    # Try 10 step using the gradient:
    niter = 0
    while True:
        f, dfdx = func(x)
        if abs(f) < tol:
            return x, niter
        if f < 0.0 and x > xmin:
            xmin = x
        elif f > 0.0 and x < xmax:
            xmax = x
        dx = -f / max(dfdx, 1e-18)
        if niter == 10 or abs(dx) > 0.3 or not (xmin < x + dx < xmax):
            break  # try bisection
        x += dx
        niter += 1

    # Bracket the solution:
    if not np.isfinite(xmin):
        xmin = x
        fmin = f
        step = 0.01
        while fmin > tol:
            xmin -= step
            fmin = func(xmin)[0]
            step *= 2

    if not np.isfinite(xmax):
        xmax = x
        fmax = f
        step = 0.01
        while fmax < 0:
            xmax += step
            fmax = func(xmax)[0]
            step *= 2

    # Bisect:
    while True:
        x = (xmin + xmax) / 2
        f = func(x)[0]
        if abs(f) < tol:
            return x, niter
        if f > 0:
            xmax = x
        else:
            xmin = x
        niter += 1
        assert niter < 1000


def collect_eigelvalues(eig_qn: np.ndarray,
                        weight_q: np.ndarray,
                        bd: BandDescriptor,
                        kpt_comm: MPICommunicator) -> Tuple[np.ndarray,
                                                            np.ndarray,
                                                            np.ndarray]:
    """Collect eigenvalues from bd.comm and kpt_comm."""
    nkpts_r = np.zeros(kpt_comm.size, int)
    nkpts_r[kpt_comm.rank] = len(weight_q)
    kpt_comm.sum(nkpts_r)
    nk = cast(int, nkpts_r.sum())
    weight_k = np.zeros(nk)
    k1 = nkpts_r[:kpt_comm.rank].sum()
    k2 = k1 + len(weight_q)
    weight_k[k1:k2] = weight_q
    kpt_comm.sum(weight_k, 0)

    eig_kn: Array2D = np.zeros((0, 0))
    k = 0
    for rank, nkpts in enumerate(nkpts_r):
        for q in range(nkpts):
            if rank == kpt_comm.rank:
                eig_n = eig_qn[q]
                eig_n = bd.collect(eig_n)
            if bd.comm.rank == 0:
                if kpt_comm.rank == 0:
                    if k == 0:
                        eig_kn = np.empty((nk, len(eig_n)))
                    if rank == 0:
                        eig_kn[k] = eig_n
                    else:
                        kpt_comm.receive(eig_kn[k], rank)
                elif rank == kpt_comm.rank:
                    kpt_comm.send(eig_n, 0)
            k += 1
    return eig_kn, weight_k, nkpts_r


def distribute_occupation_numbers(f_kn: np.ndarray,  # input
                                  f_qn: np.ndarray,  # output
                                  nkpts_r: np.ndarray,
                                  bd: BandDescriptor,
                                  kpt_comm: MPICommunicator) -> None:
    """Distribute occupation numbers over bd.comm and kpt_comm."""
    k = 0
    for rank, nkpts in enumerate(nkpts_r):
        for q in range(nkpts):
            if kpt_comm.rank == 0:
                if rank == 0:
                    if bd.comm.size == 1:
                        f_qn[q] = f_kn[k]
                    else:
                        bd.distribute(None if f_kn is None else f_kn[k],
                                      f_qn[q])
                elif f_kn is not None:
                    kpt_comm.send(f_kn[k], rank)
            elif rank == kpt_comm.rank:
                if bd.comm.size == 1:
                    kpt_comm.receive(f_qn[q], 0)
                else:
                    if bd.comm.rank == 0:
                        f_n = bd.empty(global_array=True)
                        kpt_comm.receive(f_n, 0)
                    else:
                        f_n = None
                    bd.distribute(f_n, f_qn[q])
            k += 1


class ZeroWidth(OccupationNumberCalculator):
    name = 'zero-width'
    extrapolate_factor = 0.0

    def todict(self):
        return {'width': 0.0}

    def __str__(self):
        return '# Zero width'

    def distribution(self, eig_n, fermi_level):
        f_n = np.zeros_like(eig_n)
        f_n[eig_n < fermi_level] = 1.0
        f_n[eig_n == fermi_level] = 0.5
        return f_n, np.zeros_like(eig_n), np.zeros_like(eig_n)

    def _calculate(self,
                   nelectrons,
                   eig_qn,
                   weight_q,
                   f_qn,
                   fermi_level_guess=nan):
        eig_kn, weight_k, nkpts_r = collect_eigelvalues(eig_qn, weight_q,
                                                        self.bd, self.kpt_comm)

        if eig_kn.size != 0:
            # Try to use integer weights (avoid round-off errors):
            N = int(round(1 / min(weight_k)))
            w_k = (weight_k * N).round().astype(int)
            if abs(w_k - N * weight_k).max() > 1e-10:
                # Did not work.  Use original fractional weights:
                w_k = weight_k
                N = 1

            f_kn = np.zeros_like(eig_kn)
            f_m = f_kn.ravel()
            w_kn = np.empty_like(eig_kn, dtype=w_k.dtype)
            w_kn[:] = w_k[:, np.newaxis]
            eig_m = eig_kn.ravel()
            w_m = w_kn.ravel()
            m_i = eig_m.argsort()
            w_i = w_m[m_i]
            sum_i = np.add.accumulate(w_i)
            filled_i = (sum_i <= nelectrons * N)
            i = sum(filled_i)
            f_m[m_i[:i]] = 1.0
            if i == len(m_i):
                fermi_level = inf
            else:
                extra = nelectrons * N - (sum_i[i - 1] if i > 0 else 0.0)
                if extra > 0:
                    assert extra <= w_i[i]
                    f_m[m_i[i]] = extra / w_i[i]
                    fermi_level = eig_m[m_i[i]]
                else:
                    fermi_level = (eig_m[m_i[i]] + eig_m[m_i[i - 1]]) / 2
        else:
            f_kn = None
            fermi_level = nan

        distribute_occupation_numbers(f_kn, f_qn, nkpts_r,
                                      self.bd, self.kpt_comm)

        if self.kpt_comm.rank == 0:
            fermi_level = broadcast_float(fermi_level, self.bd.comm)
        fermi_level = broadcast_float(fermi_level, self.kpt_comm)

        e_entropy = 0.0
        return fermi_level, e_entropy


class FixedOccupationNumbers(OccupationNumberCalculator):
    extrapolate_factor = 0.0

    def __init__(self, numbers, parallel_layout: ParallelLayout = None):
        """Fixed occupation numbers.

        f_sn: ndarray, shape=(nspins, nbands)
            Occupation numbers (in the range from 0 to 1)

        Example (excited state with 4 electrons)::

            occ = FixedOccupationNumbers([[1, 0, 1, 0], [1, 1, 0, 0]])

        """
        OccupationNumberCalculator.__init__(self, parallel_layout)
        self.f_sn = np.array(numbers)

    def _calculate(self,
                   nelectrons,
                   eig_qn,
                   weight_q,
                   f_qn,
                   fermi_level_guess=nan):

        calc_fixed(self.bd, self.f_sn, f_qn)

        return inf, 0.0

    def todict(self):
        return {'name': 'fixed', 'numbers': self.f_sn}


class FixedOccupationNumbersUniform(OccupationNumberCalculator):

    extrapolate_factor = 0.0
    name = 'fixed-uniform'

    def __init__(self, nelectrons, nspins, magmom, nkpts, nbands,
                 parallel_layout: ParallelLayout = None):
        """
        Uniform distribution of occupation numbers: each
        k-point per spin has the same number of occupied states
        Magnetic moment defines difference between two spins occ. numb.
        """
        OccupationNumberCalculator.__init__(self, parallel_layout)

        def get_f(nelectrons, magmom, nkpts, nbands, spin):
            """
            :param nelectrons:
            :param magmom:
            :param nkpts:
            :param nbands:
            :param spin: +1 or -1
            :return: occupation numbers per spin channel
            """

            f_qn = np.zeros(shape=(nkpts, nbands))
            nelecps = (nelectrons + spin * magmom) / 2
            assert int(nelecps) < nbands, 'need more bands!'

            f_qn[:, : int(nelecps)] = 1.0
            f_qn[:, int(nelecps)] = nelecps - int(nelecps)

            return f_qn

        if nspins == 2:
            f1_qn = get_f(nelectrons, magmom, nkpts, nbands, 1)
            f2_qn = get_f(nelectrons, magmom, nkpts, nbands, -1)
            f_qn = []
            for f1_n, f2_n in zip(f1_qn, f2_qn):
                f_qn += [f1_n, f2_n]
        else:
            f_qn = get_f(nelectrons, magmom, nkpts, nbands, 0)

        self.f_sn = np.array(f_qn)
        self.nspins = nspins
        self.magmom = magmom

    def _calculate(self,
                   nelectrons,
                   eig_qn,
                   weight_q,
                   f_qn,
                   fermi_level_guess=nan):

        calc_fixed(self.bd, self.f_sn, f_qn)

        eig_kn, weight_k, nkpts_r = collect_eigelvalues(
            eig_qn, weight_q, self.bd, self.kpt_comm)

        def get_homo(eig_kn, nelectrons, deg, magmom, spin):
            nelecps = int((nelectrons * deg + spin * magmom) / 2)
            return np.max(eig_kn[:, np.maximum(nelecps - 1, 0)])

        def get_lumo(eig_kn, nelectrons, deg, magmom, spin):
            nelecps = int((nelectrons * deg + spin * magmom) / 2)
            return np.min(eig_kn[:, nelecps])

        deg = 3 - self.nspins
        mm = self.magmom
        if eig_kn.size != 0:
            if self.nspins == 2:
                hup = get_homo(eig_kn[::2], nelectrons, deg, mm, 1)
                hdown = get_homo(eig_kn[1::2], nelectrons, deg, mm, -1)
                homo = np.maximum(hup, hdown)

                lup = get_lumo(eig_kn[::2], nelectrons, deg, mm, 1)
                ldown = get_lumo(eig_kn[1::2], nelectrons, deg, mm, -1)
                lumo = np.maximum(lup, ldown)

            else:
                homo = get_homo(eig_kn, nelectrons, deg, mm, 0)
                lumo = get_lumo(eig_kn, nelectrons, deg, mm, 0)
            fermi_level = (homo + lumo) / 2
        else:
            fermi_level = nan

        if self.kpt_comm.rank == 0:
            fermi_level = broadcast_float(fermi_level, self.bd.comm)
        fermi_level = broadcast_float(fermi_level, self.kpt_comm)

        return fermi_level, 0.0

    def todict(self):
        return {'name': 'fixed-uniform'}

    def __str__(self):
        return '# Uniform distribution of occupation numbers'


def calc_fixed(bd, f_sn, f_qn):
    if bd.nbands == f_sn.shape[1]:
        for q, f_n in enumerate(f_qn):
            s = q % len(f_sn)
            bd.distribute(f_sn[s], f_n)
    else:
        # Non-collinear calculation:
        bd.distribute(f_sn.T.flatten().copy(), f_qn[0])


def FixedOccupations(f_sn):
    warnings.warn(
        "Please use occupations={'name': 'fixed', 'numbers': ...} instead.")
    if len(f_sn) == 1:
        f_sn = np.array(f_sn) / 2
    return {'name': 'fixed', 'numbers': f_sn}


class ThomasFermiOccupations(OccupationNumberCalculator):
    name = 'orbital-free'
    extrapolate_factor = 0.0

    def _calculate(self,
                   nelectrons,
                   eig_qn,
                   weight_q,
                   f_qn,
                   fermi_level_guess=nan):
        assert len(f_qn) == 1
        f_qn[0][:] = [nelectrons]
        return inf, 0.0


def create_occ_calc(dct: Dict[str, Any],
                    *,
                    parallel_layout: ParallelLayout = None,
                    fixed_magmom_value=None,
                    rcell=None,
                    monkhorst_pack_size=None,
                    bz2ibzmap=None,
                    nspins=None,
                    nelectrons=None,
                    nkpts=None,
                    nbands=None
                    ) -> OccupationNumberCalculator:
    """Surprise: Create occupation-number object.

    The unit of width is eV and name must be one of:

    * 'fermi-dirac'
    * 'marzari-vanderbilt'
    * 'methfessel-paxton'
    * 'fixed'
    * 'tetrahedron-method'
    * 'improved-tetrahedron-method'
    * 'orbital-free'

    >>> occ = create_occ_calc({'width': 0.0})
    >>> occ.calculate(nelectrons=3,
    ...               eigenvalues=[[0, 1, 2], [0, 2, 3]],
    ...               weights=[1, 1])
    (array([[1., 1., 0.],
           [1., 0., 0.]]), [1.5], 0.0)
    """
    kwargs = dct.copy()
    fix_the_magnetic_moment = kwargs.pop('fixmagmom', False)
    name = kwargs.pop('name', '')
    kwargs['parallel_layout'] = parallel_layout

    if name == 'unknown':
        return OccupationNumberCalculator(**kwargs)

    occ: OccupationNumberCalculator

    if kwargs.get('width') == 0.0:
        del kwargs['width']
        occ = ZeroWidth(**kwargs)
    elif name == 'methfessel-paxton':
        occ = MethfesselPaxtonCalculator(**kwargs)
    elif name == 'fermi-dirac':
        occ = FermiDiracCalculator(**kwargs)
    elif name == 'marzari-vanderbilt':
        occ = MarzariVanderbiltCalculator(**kwargs)
    elif name in {'tetrahedron-method', 'improved-tetrahedron-method'}:
        from gpaw.tetrahedron import TetrahedronMethod
        occ = TetrahedronMethod(rcell,
                                monkhorst_pack_size,
                                name == 'improved-tetrahedron-method',
                                bz2ibzmap,
                                **kwargs)
    elif name == 'orbital-free':
        return ThomasFermiOccupations(**kwargs)
    elif name == 'fixed':
        return FixedOccupationNumbers(**kwargs)
    elif name == 'fixed-uniform':
        return FixedOccupationNumbersUniform(
            nelectrons, nspins, fixed_magmom_value,
            nkpts, nbands, **kwargs)
    else:
        raise ValueError(f'Unknown occupation number object name: {name}')

    if fix_the_magnetic_moment:
        occ = FixMagneticMomentOccupationNumberCalculator(
            occ, fixed_magmom_value)

    return occ


def occupation_numbers(occ, eig_skn, weight_k, nelectrons):
    """Calculate occupation numbers from eigenvalues in eV (**deprecated**).

    occ: dict
        Example: {'name': 'fermi-dirac', 'width': 0.05} (width in eV).
    eps_skn: ndarray, shape=(nspins, nibzkpts, nbands)
        Eigenvalues.
    weight_k: ndarray, shape=(nibzkpts,)
        Weights of k-points in IBZ (must sum to 1).
    nelectrons: int or float
        Number of electrons.

    Returns a tuple containing:

    * f_skn (sums to nelectrons)
    * fermi-level [Hartree]
    * magnetic moment
    * entropy as -S*T [Hartree]
    """

    warnings.warn('Please use one of the OccupationNumbers implementations',
                  DeprecationWarning)
    occ = create_occ_calc(occ)
    f_kn, (fermi_level,), e_entropy = occ.calculate(
        nelectrons * len(eig_skn) / 2,
        [eig_n for eig_kn in eig_skn for eig_n in eig_kn],
        list(weight_k) * len(eig_skn))

    f_kn *= np.array(weight_k)[:, np.newaxis]

    if len(eig_skn) == 1:
        f_skn = np.array([f_kn]) * 2
        e_entropy *= 2
        magmom = 0.0
    else:
        f_skn = np.array([f_kn[::2], f_kn[1::2]])
        f1, f2 = f_skn.sum(axis=(1, 2))
        magmom = f1 - f2

    return f_skn, fermi_level * Ha, magmom, e_entropy * Ha


def FermiDirac(width, fixmagmom=False):
    return dict(name='fermi-dirac', width=width, fixmagmom=fixmagmom)


def MarzariVanderbilt(width, fixmagmom=False):
    return dict(name='marzari-vanderbilt', width=width, fixmagmom=fixmagmom)


def MethfesselPaxton(width, order=0, fixmagmom=False):
    return dict(name='methfessel-paxton', width=width, order=order,
                fixmagmom=fixmagmom)



"""
The solvated jellium method is contained in this module.
This enables electronically grand-canonical calculations to be calculated,
typically for simulating electrochemical interfaces.
"""

import os
import copy
import textwrap

import numpy as np
from scipy.stats import linregress
import ase.io
from ase.units import Bohr, Ha
from ase.calculators.calculator import (Parameters, equal, InputError,
                                        PropertyNotPresent)
from ase.dft.bandgap import bandgap
from ase.parallel import paropen

import gpaw.mpi
from gpaw import ConvergenceError
from gpaw.jellium import Jellium, JelliumSlab
from gpaw.hamiltonian import RealSpaceHamiltonian
from gpaw.fd_operators import Gradient
from gpaw.dipole_correction import DipoleCorrection
from gpaw.solvation.cavity import Power12Potential, get_pbc_positions
from gpaw.solvation.calculator import SolvationGPAW
from gpaw.solvation.hamiltonian import SolvationRealSpaceHamiltonian
from gpaw.solvation.poisson import WeightedFDPoissonSolver


class SJM(SolvationGPAW):
    r"""Solvated Jellium method.
    (Implemented as a subclass of the SolvationGPAW class.)

    The method allows the simulation of an electrochemical environment, where
    the potential can be varied by changing the charging (that is, number of
    electrons) of the system. For this purpose, it allows the usagelof non-
    neutral periodic slab systems. Cell neutrality is achieved by adding a
    background charge in the solvent region above the slab

    Further details are given in https://doi.org/10.1021/acs.jpcc.8b02465
    If you use this method, we appreciate it if you cite that work.

    The method can be run in two modes:

        - Constant charge: The number of excess electrons in the simulation
          can be directly specified with the 'excess_electrons' keyword,
          leaving 'target_potential' set to None.
        - Constant potential: The target potential (expressed as a work
          function) can be specified with the 'target_potential' keyword.
          Optionally, the 'excess_electrons' keyword can be supplied to specify
          the initial guess of the number of electrons.

    By default, this method writes the grand-potential energy to the output;
    that is, the energy that has been adjusted with `- \mu N` (in this case,
    `\mu` is the work function and `N` is the excess electrons). This is the
    energy that is compatible with the forces in constant-potential mode and
    thus will behave well with optimizers, NEBs, etc. It is also frequently
    used in subsequent free-energy calculations.

    Within this method, the potential is expressed as the top-side work
    function of the slab. Therefore, a potential of 0 V_SHE corresponds to
    a work function of roughly 4.4 eV. (That is, the user should specify
    target_potential as 4.4 in this case.) Because this method is
    attempting to bring the work function to a target value, the work
    function itself needs to be well-converged. For this reason, the
    'work function' keyword is automatically added to the SCF convergence
    dictionary with a value of 0.001. This can be overriden by the user.

    This method requires a dipole correction, and this is turned on
    automatically, but can be overridden with the poissonsolver keyword.

    The SJM class takes a single argument, the sj dictionary. All other
    arguments are fed to the parent SolvationGPAW (and therefore GPAW)
    calculator.

    Parameters:

    sj: dict
        Dictionary of parameters for the solvated jellium method, whose
        possible keys are given below.

    Parameters in sj dictionary:

    excess_electrons: float
        Number of electrons added in the atomic system and (with opposite
        sign) in the background charge region. If the 'target_potential'
        keyword is also supplied, 'excess_electrons' is taken as an initial
        guess for the needed number of electrons.
    target_potential: float
        The potential that should be reached or kept in the course of the
        calculation. If set to 'None' (default) a constant-charge
        calculation based on the value of 'excess_electrons' is performed.
        Expressed as a work function, on the top side of the slab; see note
        above.
    tol: float
        Tolerance for the deviation of the target potential.
        If the potential is outside the defined range 'ne' will be
        changed in order to get inside again. Default: 0.01 V.
    max_iters: int
        In constant-potential mode, the maximum number of iterations to try
        to equilibrate the potential (by varying ne). Default: 10.
    jelliumregion: dict
        Parameters regarding the shape of the counter charge region.
        Implemented keys:

        'top': float or None
            Upper boundary of the counter charge region (z coordinate).
            A positive float is taken as the upper boundary coordiate.
            A negative float is taken as the distance below the uppper
            cell boundary.  Default: -1.
        'bottom': float or 'cavity_like' or None
            Lower boundary of the counter-charge region (z coordinate).
            A positive float is taken as the lower boundary coordinate.
            A negative float is interpreted as the distance above the highest
            z coordinate of any atom; this can be fixed by setting
            fix_bottom to True. If 'cavity_like' is given the counter
            charge will take the form of the cavity up to the 'top'.
            Default: -3.
        'thickness': float or None
            Thickness of the counter charge region in Angstroms.
            Can only be used if start is not 'cavity_like'. Default: None
        'fix_bottom': bool
            Do not allow the lower limit of the jellium region to adapt
            to a change in the atomic geometry during a relaxation.
            Omitted in a 'cavity_like' jelliumregion.
            Default: False
    grand_output: bool
        Write the grand-potential energy into output files such as
        trajectory files. Default: True
    always_adjust: bool
        Adjust ne again even when potential is within tolerance.
        This is useful to set to True along with a loose potential
        tolerance (tol) to allow the potential and structure to be
        simultaneously optimized in a geometry optimization, for example.
        Default: False.
    slope : float or None
        Initial guess of the slope, in volts per electron, of the relationship
        between the electrode potential and excess electrons, used to
        equilibrate the potential.
        This only applies to the first step, as the slope is calculated
        internally at subsequent steps. If None, will be guessed based on
        apparent capacitance of 10 uF/cm^2.
    max_step : float
        When equilibrating the potential, if a step results in the
        potential being further from the target and changing by more than
        'max_step' V, then the step size is cut in half and we try again. This
        is to avoid leaving the linear region. You can set to np.inf to turn
        this off. Default: 2 V.
    mixer : float
        Damping for the slope estimate. Because estimating slopes can
        sometimes be noisy (particularly for small changes, or when
        positions have also changed), some fraction of the previous slope
        estimate can be "mixed" with the current slope estimate before the
        new slope is established. E.g., new_slope = mixer * old_slope +
        (1. - mixer) * current_slope_estimate. Set to 0 for no damping.
        Default: 0.5.

    Special SJM methods (in addition to those of GPAW/SolvationGPAW):

    get_electrode_potential
        Returns the potential of the simulated electrode, in V, relative
        to the vacuum.

    write_sjm_traces
        Write traces of quantities like electrostatic potential or cavity
        to disk.

    """
    implemented_properties = ['energy', 'forces', 'stress', 'dipole',
                              'magmom', 'magmoms', 'excess_electrons',
                              'electrode_potential']
    _sj_default_parameters = Parameters(
        {'excess_electrons': 0.,
         'jelliumregion': {'top': -1.,
                           'bottom': -3.,
                           'thickness': None,
                           'fix_bottom': False},
         'target_potential': None,
         'tol': 0.01,
         'always_adjust': False,
         'grand_output': True,
         'max_iters': 10,
         'max_step': 2.,
         'slope': None,
         'mixer': 0.5})
    default_parameters = copy.deepcopy(SolvationGPAW.default_parameters)
    default_parameters.update({'poissonsolver': {'dipolelayer': 'xy'}})
    default_parameters['convergence'].update({'work function': 0.001})
    default_parameters.update({'sj': _sj_default_parameters})

    def __init__(self, restart=None, **kwargs):

        deprecated_keys = ['ne', 'potential', 'write_grandcanonical_energy',
                           'potential_equilibration_mode', 'dpot',
                           'max_pot_deviation', 'doublelayer', 'verbose']
        msg = ('{:s} is no longer a supported keyword argument for the SJM '
               'class. All SJM arguments should be sent in via the "sj" '
               'dict.')
        for key in deprecated_keys:
            if key in kwargs:
                raise InputError(textwrap.fill(msg.format(key)))

        # Note the below line calls self.set().
        SolvationGPAW.__init__(self, restart, **kwargs)

    def set(self, **kwargs):
        """Change parameters for calculator.

        It differs from the standard `set` function in two ways:
        - Keywords in the `sj` dictionary are handled.
        - It does not reinitialize and delete `self.wfs` if only the
          background charge is changed.

        """
        p = self.parameters['sj']

        # We handle 'sj' and 'background_charge' internally; passing to GPAW's
        # set function will trigger a deletion of the density, etc.
        sj_changes = kwargs.pop('sj', {})
        try:
            sj_changes = {key: value for key, value in sj_changes.items()
                          if not equal(value, p[key])}
        except KeyError:
            raise InputError(
                'Unexpected key(s) provided to sj dict. '
                'Keys provided were "{}". '
                'Only keys allowed are "{}".'
                .format(', '.join(sj_changes),
                        ', '.join(self.default_parameters['sj'])))
        p.update(sj_changes)
        background_charge = kwargs.pop('background_charge', None)

        SolvationGPAW.set(self, **kwargs)

        # parent_changed checks if GPAW needs to be reinitialized
        # The following key do not need reinitialization
        parent_changed = False
        for key in kwargs:
            if key not in ['mixer', 'verbose', 'txt', 'hund', 'random',
                           'eigensolver', 'convergence', 'fixdensity',
                           'maxiter']:
                parent_changed = True

        if len(sj_changes):
            if self.wfs is None:
                self.log('Non-default Solvated Jellium parameters:')
            else:
                self.log('Changed Solvated Jellium parameters:')
            self.log.print_dict({i: p[i] for i in sj_changes})
            self.log()

        if 'target_potential' in sj_changes and p.target_potential is not None:
            # If target potential is changed by the user and the slope is
            # known, a step towards the new potential is taken right away.
            try:
                true_potential = self.get_electrode_potential()
            # TypeError is needed for the case of starting from a gpw
            # file and changing the target potential at the start.
            except (AttributeError, TypeError):
                pass
            else:
                if self.atoms and p.slope:
                    p.excess_electrons += ((p.target_potential -
                                            true_potential) / p.slope)
                    self.log('Number of electrons changed to {:.4f} based '
                             'on slope of {:.4f} V/electron.'
                             .format(p.excess_electrons, p.slope))

        if (any(key in ['target_potential', 'excess_electrons',
            'jelliumregion'] for key in sj_changes) and not parent_changed):
            self.results = {}
            # SolvationGPAW will not reinitialize anymore if only
            # 'sj' keywords are set. The lines below will reinitialize and
            # apply the changed charges.
            if self.atoms:
                self.set(background_charge=self._create_jellium())

        if 'tol' in sj_changes:
            try:
                true_potential = self.get_electrode_potential()
            except (AttributeError, TypeError):
                pass
            else:
                msg = ('Potential tolerance changed to {:1.4f} V: '
                       .format(p.tol))
                if abs(true_potential - p.target_potential) > p.tol:
                    msg += 'new calculation required.'
                    self.results = {}
                    if self.atoms and p.slope is not None:
                        p.excess_electrons += ((p.target_potential -
                                               true_potential) / p.slope)
                        self.set(background_charge=self._create_jellium())
                        msg += ('\n Excess electrons changed to {:.4f} based '
                                'on slope of {:.4f} V/electron.'
                                .format(p.excess_electrons, p.slope))
                else:
                    msg += 'already within tolerance.'
                self.log(msg)

        if background_charge:
            # background_charge is a GPAW parameter that we handle internally,
            # as it contains the jellium countercharge. Note if a user tries to
            # specify an *additional* background charge this will probably
            # conflict, but we know of no such use cases.
            if self.wfs is None:
                kwargs.update({'background_charge': background_charge})
                SolvationGPAW.set(self, **kwargs)
            else:
                if parent_changed:
                    self.density = None
                else:
                    if self.density.background_charge:
                        self.density.background_charge = background_charge
                        self.density.background_charge.set_grid_descriptor(
                            self.density.finegd)
                    self._quick_reinitialization()

                self.wfs.nvalence = self.setups.nvalence + p.excess_electrons
                self.log('Number of valence electrons is now {:.5f}'
                         .format(self.wfs.nvalence))

    def _quick_reinitialization(self):
        """Minimal reinitialization of electronic-structure stuff when only
        background charge changes."""
        if self.density.nct_G is None:
            self.initialize_positions()

        self.density.reset()
        self._set_atoms(self.atoms)
        self.density.mixer.reset()
        self.wfs.initialize(self.density, self.hamiltonian,
                            self.spos_ac)
        self.wfs.eigensolver.reset()
        if self.scf:
            self.scf.reset()

    def calculate(self, atoms=None, properties=['energy'],
                  system_changes=['cell']):
        """Perform an electronic structure calculation, with either a
        constant number of electrons or a target potential, as requested by
        the user in the 'sj' dict."""

        if atoms and not self.atoms:
            # Need to be set before ASE's Calculator.calculate gets to it.
            self.atoms = atoms.copy()

        if len(system_changes) == 0 and len(self.results) > 0:
            # Potential is already equilibrated.
            SolvationGPAW.calculate(self, atoms, properties, system_changes)
            return

        self.log('Solvated jellium method (SJM) calculation:')

        p = self.parameters['sj']

        if p.target_potential is None:
            self.log('Constant-charge calculation with {:.5f} excess '
                     'electrons'.format(p.excess_electrons))
            # Background charge is set here, not earlier, because atoms needed.
            self.set(background_charge=self._create_jellium())
            SolvationGPAW.calculate(self, atoms, ['energy'], system_changes)
            self.log('Potential found to be {:.5f} V (with {:+.5f} '
                     'electrons)'.format(self.get_electrode_potential(),
                                         p.excess_electrons))
        else:
            self.log(' Constant-potential mode.')
            self.log(' Target potential: {:.5f} +/- {:.5f}'
                     .format(p.target_potential, p.tol))
            self.log(' Initial guess of excess electrons: {:.5f}'
                     .format(p.excess_electrons))
            if 'workfunction' in self.parameters.convergence:
                if self.parameters.convergence['workfunction'] >= p.tol:
                    msg = ('Warning: it appears that your work function '
                           'convergence criterion ({:g}) is higher than your '
                           'desired potential tolerance ({:g}). This may lead '
                           'to issues with potential convergence.'
                           .format(self.parameters.convergence['workfunction'],
                                   p.tol))
                    self.log(textwrap.fill(msg))
            self._equilibrate_potential(atoms, system_changes)
        if properties != ['energy']:
            # The equilibration loop only calculated energy, to save
            # unnecessary computations (mostly of forces) in the loop.
            SolvationGPAW.calculate(self, atoms, properties, [])

        # Note that grand-potential energies were assembled in summary,
        # which in turn was called by GPAW.calculate.

        if p.grand_output:
            self.results['energy'] = self.omega_extrapolated * Ha
            self.results['free_energy'] = self.omega_free * Ha
            self.log('Grand-potential energy was written into results.\n')
        else:
            self.log('Canonical energy was written into results.\n')

        self.results['excess_electrons'] = p.excess_electrons
        self.results['electrode_potential'] = self.get_electrode_potential()
        self.log.fd.flush()

    def _equilibrate_potential(self, atoms, system_changes):
        """Adjusts the number of electrons until the potential reaches the
        desired value."""
        p = self.parameters['sj']
        iteration = 0
        previous_electrons = []
        previous_potentials = []

        rerun = False
        while iteration <= p.max_iters:
            self.log('Attempt {:d} to equilibrate potential to {:.3f} +/-'
                     ' {:.3f} V'
                     .format(iteration, p.target_potential, p.tol))
            self.log('Current guess of excess electrons: {:+.5f}\n'
                     .format(p.excess_electrons))
            if iteration == 1:
                self.timer.start('Potential equilibration loop')
                # We don't want SolvationGPAW to see any more system
                # changes, like positions, after attempt 0.
                system_changes = []

            if any([iteration, rerun, 'positions' in system_changes]):
                self.set(background_charge=self._create_jellium())

            # Do the calculation.
            SolvationGPAW.calculate(self, atoms, ['energy'], system_changes)
            true_potential = self.get_electrode_potential()
            self.log()
            msg = (f'Potential found to be {true_potential:.5f} V (with '
                   f'{p.excess_electrons:+.5f} excess electrons, attempt '
                   f'{iteration:d}/{p.max_iters:d}')
            msg += ' rerun).' if rerun else ').'
            self.log(msg, flush=True)

            # Check if we took too big of a step.
            try:
                stepsize = abs(true_potential - previous_potentials[-1])
            except IndexError:
                pass
            else:
                diff_ratio = (true_potential - p.target_potential) /\
                    (previous_potentials[-1] - p.target_potential)
                if stepsize > p.max_step and diff_ratio < -0.5:
                    self.log('Step resulted in a potential change of '
                             f'{stepsize:.2f} V, larger than max_step '
                             f'({p.max_step:.2f} V) and\n surpassed the'
                             ' target potential by a dangerous amount.\n'
                             ' The step is rejected and the change in'
                             ' excess_electrons will be halved.')
                    p.excess_electrons = (previous_electrons[-1] +
                                          (p.excess_electrons -
                                           previous_electrons[-1]) * 0.5)
                    rerun = True
                    continue  # back to while

            # Increase iteration count.
            iteration += 1
            rerun = False

            # Store attempt and calculate slope.
            previous_electrons.append(float(p.excess_electrons))
            previous_potentials.append(float(true_potential))
            if len(previous_electrons) > 1:
                slope = _calculate_slope(previous_electrons,
                                         previous_potentials)
                self.log('Slope regressed from last {:d} attempts is '
                         '{:.4f} V/electron,'
                         .format(len(previous_electrons[-4:]), slope))
                area = np.product(np.diag(atoms.cell[:2, :2]))
                capacitance = -1.6022 * 1e3 / (area * slope)
                self.log(f'or apparent capacitance of {capacitance:.4f} '
                         'muF/cm^2')
                if p.slope is not None:
                    p.slope = p.mixer * p.slope + (1. - p.mixer) * slope
                    self.log(f'After mixing with {p.mixer:.2f}, new slope is '
                             f'{p.slope:.4f} V/electron.')
                else:
                    p.slope = slope
                self.log.flush()

            # Check if we're equilibrated and exit if always_adjust is False.
            if abs(true_potential - p.target_potential) < p.tol:
                self.log('Potential is within tolerance. Equilibrated.')
                if iteration >= 2:
                    self.timer.stop('Potential equilibration loop')
                if not p.always_adjust:
                    return

            # Guess slope if we don't have enough information yet.
            if p.slope is None:
                area = np.product(np.diag(atoms.cell[:2, :2]))
                p.slope = -1.6022e3 / (area * 10.)
                self.log('No slope provided, guessing a slope of '
                         f'{p.slope:.4f} corresponding\nto an apparent '
                         'capacitance of 10 muF/cm^2.')

            # Finally, update the number of electrons.
            p.excess_electrons += ((p.target_potential - true_potential)
                                   / p.slope)
            self.log(f'Number of electrons changed to {p.excess_electrons:.4f}'
                     f' based on slope of {p.slope:.4f} V/electron.')

            # Check if we're equilibrated and exit if always_adjust is True.
            if (abs(true_potential - p.target_potential) < p.tol
                and p.always_adjust):
                return

        msg = (f'Potential could not be reached after {iteration-1:d} '
               'iterations. This may indicate your workfunction is noisier '
               'than your potential tol. You may try setting the '
               'convergence["workfunction"] keyword. The last values of '
               'excess_electrons and the potential are listed below; '
               'plotting them could give you insight into the problem.')
        msg = textwrap.fill(msg) + '\n'
        for n, p in zip(previous_electrons, previous_potentials):
            msg += '{:+.6f} {:.6f}\n'.format(n, p)
        self.log(msg, flush=True)
        raise PotentialConvergenceError(msg)

    def write_sjm_traces(self, path='sjm_traces', style='z',
                         props=('potential', 'cavity', 'background_charge')):
        """Write traces of quantities in `props` to file on disk; traces will
        be stored within specified path. Default is to save as vertical traces
        (style 'z'), but can also save as cube (specify `style='cube'`)."""
        grid = self.density.finegd
        data = {'cavity': self.hamiltonian.cavity.g_g,
                'background_charge': self.density.background_charge.mask_g,
                'potential': (self.hamiltonian.vHt_g * Ha -
                              self.get_fermi_level())}
        if not os.path.exists(path) and gpaw.mpi.world.rank == 0:
            os.makedirs(path)
        for prop in props:
            if style == 'z':
                _write_trace_in_z(grid, data[prop], prop + '.txt', path)
            elif style == 'cube':
                _write_property_on_grid(grid, data[prop], self.atoms,
                                        prop + '.cube', path)

    def summary(self):
        """Writes summary information to the log file.
        This varies from the implementation in gpaw.calculator.GPAW by the
        inclusion of grand potential quantities."""
        # Standard GPAW summary.
        self.hamiltonian.summary(self.wfs, self.log)
        # Add grand-canonical terms.
        p = self.parameters['sj']
        self.log()
        mu_N = self.get_electrode_potential() * p.excess_electrons / Ha
        self.omega_free = self.hamiltonian.e_total_free + mu_N
        self.omega_extrapolated = self.hamiltonian.e_total_extrapolated + mu_N
        self.log('Legendre-transformed energies (grand potential, '
                 'Omega = E - N mu)')
        self.log(' N (excess electrons):  {:+11.6f}'
                 .format(p.excess_electrons))
        self.log(' mu (workfunction, eV): {:+11.6f}'
                 .format(self.get_electrode_potential()))
        self.log(' (Grand) free energy:   {:+11.6f}'
                 .format(Ha * self.omega_free))
        self.log(' (Grand) extrapolated:  {:+11.6f}'
                 .format(Ha * self.omega_extrapolated))
        self.log()
        # Back to standard GPAW summary.
        self.density.summary(self.atoms, self.results.get('magmom', 0.0),
                             self.log)
        self.wfs.summary(self.log)
        if len(self.wfs.fermi_levels) == 1:
            try:
                bandgap(self,
                        output=self.log.fd,
                        efermi=self.wfs.fermi_level * Ha)
            except ValueError:
                pass
        self.log.fd.flush()

    def _create_jellium(self):
        """Creates the counter charge according to the user's specs."""
        atoms = self.atoms

        p = self.parameters['sj']
        jellium = p['jelliumregion']
        defaults = self.default_parameters['sj']['jelliumregion']

        # Populate missing keywords.
        missing = {'top': None,
                   'bottom': None,
                   'thickness': None,
                   'fix_bottom': False}
        for key in missing:
            if key not in jellium:
                jellium[key] = missing[key]

        # Catch incompatible specifications.
        if jellium.get('thickness') and jellium['bottom'] == 'cavity_like':
            raise InputError("With a cavity-like counter charge only the "
                             "keyword 'top' (not 'thickness') allowed.")

        # We need 2 of the 3 "specifieds" below.
        specifieds = [jellium['top'] is not None,
                      jellium['bottom'] is not None,
                      jellium['thickness'] is not None]

        if sum(specifieds) == 3:
            raise InputError('The jellium region has been overspecified.')
        if sum(specifieds) == 0:
            top = defaults['top']
            bottom = defaults['bottom']
            thickness = defaults['thickness']
        if sum(specifieds) == 2:
            top = jellium['top']
            bottom = jellium['bottom']
            thickness = jellium['thickness']
        if specifieds == [True, False, False]:
            top = jellium['top']
            bottom = defaults['bottom']
            thickness = None
        elif specifieds == [False, True, False]:
            top = defaults['top']
            bottom = jellium['bottom']
            thickness = None
        elif specifieds == [False, False, True]:
            top = None
            bottom = defaults['bottom']
            thickness = jellium['thickness']

        # Deal with negative specifications of upper and lower limits;
        # as well as fixed/free lower limit.
        if top is not None and top < 0.:
            top = atoms.cell[2][2] + top
        if bottom not in [None, 'cavity_like'] and bottom < 0.:
            bottom = (max(atoms.positions[:, 2]) - bottom)
            if jellium['fix_bottom'] is True:
                try:
                    bottom = self._fixed_lower_limit
                except AttributeError:
                    self._fixed_lower_limit = bottom

        # Use thickness if needed.
        if top is None:
            top = bottom + thickness
        if bottom is None:
            bottom = top - thickness

        # Catch unphysical limits.
        if top > atoms.cell[2][2]:
            raise InputError('The upper limit of the jellium region lies '
                             'outside of unit cell. If you did not set it '
                             'manually, increase your unit cell size or '
                             'translate the atomic system down along the '
                             'z-axis.')
        if bottom != 'cavity_like':
            if bottom > top:
                raise InputError('Your jellium region has a bottom at {:.3f}'
                                 ' AA, which is above the top at {:.3f} AA.'
                                 .format(bottom, top))

        # Finally, make the jellium.
        if bottom == 'cavity_like':
            if self.hamiltonian is None:
                self.initialize(atoms)
                self.set_positions(atoms)
                g_g = self.hamiltonian.cavity.g_g.copy()
                self.wfs = None
                self.density = None
                self.hamiltonian = None
                self.initialized = False
            else:
                self.set_positions(atoms)
                # XXX If you start with a fixed bottom, run a calc,
                # then hot-switch to cavity_like it crashes here. Not sure
                # that's common enough to worry about.
                g_g = self.hamiltonian.cavity.g_g
            self.log('Jellium counter-charge defined with:\n'
                     ' Bottom: cavity-like\n'
                     ' Top: {:7.3f} AA\n'
                     ' Charge: {:5.4f}\n'
                     .format(top, p.excess_electrons))
            return CavityShapedJellium(charge=p.excess_electrons,
                                       g_g=g_g,
                                       z2=top)

        self.log('Jellium counter-charge defined with:\n'
                 ' Bottom: {:7.3f} AA\n'
                 ' Top: {:7.3f} AA\n'
                 ' Charge: {:5.4f}\n'
                 .format(bottom, top, p.excess_electrons))
        return JelliumSlab(charge=p.excess_electrons,
                           z1=bottom,
                           z2=top)

    def get_electrode_potential(self):
        """Returns the potential of the simulated electrode, in V, relative
        to the vacuum. This comes directly from the work function."""
        try:
            return Ha * self.hamiltonian.get_workfunctions(self.wfs)[1]
        except TypeError:
            # Error happens on freshly-opened *.gpw file.
            if 'electrode_potential' in self.results:
                return self.results['electrode_potential']
            else:
                msg = ('Electrode potential could not be read. Make sure a DFT'
                       ' calculation has been performed before reading the '
                       'potential.')
                raise PropertyNotPresent(textwrap.fill(msg))

    def initialize(self, atoms=None, reading=False):
        """Inexpensive initialization.

        This catches CavityShapedJellium, which GPAW's initialize does not
        know how to handle on restart. We delete and it will be recreated by
        the _create_jellium method when needed.
        """
        background_charge = self.parameters['background_charge']
        if isinstance(background_charge, dict):
            if 'z1' in background_charge:
                if background_charge['z1'] == 'cavity_like':
                    self.parameters['background_charge'] = None
        SolvationGPAW.initialize(self=self, atoms=atoms, reading=reading)

    def create_hamiltonian(self, realspace, mode, xc):
        """This differs from SolvationGPAW's create_hamiltonian method by the
        ability to use dipole corrections."""
        if not realspace:
            raise NotImplementedError(
                'SJM does not support calculations in reciprocal space yet'
                ' due to a lack of an implicit solvent module.')

        dens = self.density

        self.hamiltonian = SJM_RealSpaceHamiltonian(
            *self.stuff_for_hamiltonian,
            gd=dens.gd, finegd=dens.finegd,
            nspins=dens.nspins,
            collinear=dens.collinear,
            setups=dens.setups,
            timer=self.timer,
            xc=xc,
            world=self.world,
            redistributor=dens.redistributor,
            vext=self.parameters.external,
            psolver=self.parameters.poissonsolver,
            stencil=mode.interpolation)

        xc.set_grid_descriptor(self.hamiltonian.finegd)


def _write_trace_in_z(grid, property, name, dir):
    """Writes out a property (like electrostatic potential, cavity, or
    background charge) as a function of the z coordinate only. `grid` is the
    grid descriptor, typically self.density.finegd. `property` is the property
    to be output, on the same grid."""
    property = grid.collect(property, broadcast=True)
    property_z = property.mean(0).mean(0)
    with paropen(os.path.join(dir, name), 'w') as f:
        for i, val in enumerate(property_z):
            f.write('%f %1.8f\n' % ((i + 1) * grid.h_cv[2][2] * Bohr, val))


def _write_property_on_grid(grid, property, atoms, name, dir):
    """Writes out a property (like electrostatic potential, cavity, or
    background charge) on the grid, as a cube file. `grid` is the
    grid descriptor, typically self.density.finegd. `property` is the property
    to be output, on the same grid."""
    property = grid.collect(property, broadcast=True)
    ase.io.write(os.path.join(dir, name), atoms, data=property)


def _calculate_slope(previous_electrons, previous_potentials):
    """Calculates the slope of potential versus number of electrons;
    regresses based on (up to) last four data points to smooth noise."""
    npoints = 4
    ans = linregress(previous_electrons[-npoints:],
                     previous_potentials[-npoints:])
    return ans[0]


class SJMPower12Potential(Power12Potential):
    r"""Inverse power-law potential.
    Inverse power law potential for SJM, inherited from the
    Power12Potential of gpaw.solvation. This is a 1/r^{12} repulive
    potential taking the value u0 at the atomic radius. In SJM one also has the
    option of removing the solvent from the electrode backside and adding
    ghost plane/atoms to remove the solvent from the electrode-water interface.

    Parameters:

    atomic_radii: function
        Callable mapping an ase.Atoms object to an iterable of atomic radii
        in Angstroms. If not provided, defaults to van der Waals radii.
    u0: float
        Strength of the potential at the atomic radius in eV.
        Defaults to 0.18 eV, the best-fit value for water from Held &
        Walter.
    pbc_cutoff: float
        Cutoff in eV for including neighbor cells in a calculation with
        periodic boundary conditions.
    H2O_layer: bool, int or str
        True: Exclude the implicit solvent from the interface region
        between electrode and water. Ghost atoms will be added below
        the water layer.
        False: The opposite of True. [default]
        int: Explicitly account for the given number of water molecules above
        electrode. This is handy if H2O is directly adsorbed and a water layer
        is present in the unit cell at the same time.
        'plane': Use a plane instead of ghost atoms for freeing the surface.
    unsolv_backside: bool
        Exclude implicit solvent from the region behind the electrode

    """

    depends_on_el_density = False
    depends_on_atomic_positions = True

    def __init__(self, atomic_radii=None, u0=0.180, pbc_cutoff=1e-6,
                 tiny=1e-10, H2O_layer=False, unsolv_backside=True):
        super().__init__(atomic_radii, u0, pbc_cutoff, tiny)
        self.H2O_layer = H2O_layer
        self.unsolv_backside = unsolv_backside

    def write(self, writer):
        writer.write(
            name='SJMPower12Potential',
            u0=self.u0,
            atomic_radii=self.atomic_radii_output,
            H2O_layer=self.H2O_layer,
            unsolv_backside=self.unsolv_backside)

    def update(self, atoms, density):
        if atoms is None:
            return False
        self.r12_a = (self.atomic_radii_output / Bohr) ** 12
        r_cutoff = (self.r12_a.max() * self.u0 / self.pbc_cutoff) ** (1. / 12.)
        self.pos_aav = get_pbc_positions(atoms, r_cutoff)
        self.u_g.fill(.0)
        self.grad_u_vg.fill(.0)
        na = np.newaxis

        if self.unsolv_backside:
            # Removing solvent from electrode backside
            for z in range(self.u_g.shape[2]):
                if (self.r_vg[2, 0, 0, z] - atoms.positions[:, 2].min() /
                        Bohr < 0):
                    self.u_g[:, :, z] = np.inf
                    self.grad_u_vg[:, :, :, z] = 0

        if self.H2O_layer:
            # Add ghost coordinates and indices to pos_aav dictionary if
            # a water layer is present.

            all_oxygen_ind = [atom.index for atom in atoms
                              if atom.symbol == 'O']

            # Disregard oxygens that don't belong to the water layer
            allwater_oxygen_ind = []
            for ox in all_oxygen_ind:
                nH = 0

                for i, atm in enumerate(atoms):
                    for period_atm in self.pos_aav[i]:
                        dist = period_atm * Bohr - atoms[ox].position
                        if np.linalg.norm(dist) < 1.3 and atm.symbol == 'H':
                            nH += 1

                if nH >= 2:
                    allwater_oxygen_ind.append(ox)

            # If the number of waters in the water layer is given as an input
            # (H2O_layer=i) then only the uppermost i water molecules are
            # regarded for unsolvating the interface (this is relevant if
            # water is adsorbed on the surface)
            if not isinstance(self.H2O_layer, (bool, str)):
                if self.H2O_layer % 1 < self.tiny:
                    self.H2O_layer = int(self.H2O_layer)
                else:
                    raise InputError('Only an integer number of water '
                                     'molecules is possible in the water '
                                     'layer')

                allwaters = atoms[allwater_oxygen_ind]
                indizes_water_ox_ind = np.argsort(allwaters.positions[:, 2],
                                                  axis=0)

                water_oxygen_ind = []
                for i in range(self.H2O_layer):
                    water_oxygen_ind.append(
                        allwater_oxygen_ind[indizes_water_ox_ind[-1 - i]])

            else:
                water_oxygen_ind = allwater_oxygen_ind

            oxygen = self.pos_aav[water_oxygen_ind[0]] * Bohr
            if len(water_oxygen_ind) > 1:
                for windex in water_oxygen_ind[1:]:
                    oxygen = np.concatenate(
                        (oxygen, self.pos_aav[windex] * Bohr))

            O_layer = []
            if isinstance(self.H2O_layer, str):
                # Add a virtual plane
                if len(self.H2O_layer.split('-')) > 1:
                    plane_z = float(self.H2O_layer.split('-')[1]) - \
                        1.0 * self.atomic_radii_output[water_oxygen_ind[0]]
                else:
                    plane_rel_oxygen = -1.5 * self.atomic_radii_output[
                        water_oxygen_ind[0]]
                    plane_z = oxygen[:, 2].min() + plane_rel_oxygen

                r_diff_zg = self.r_vg[2, :, :, :] - plane_z / Bohr
                r_diff_zg[r_diff_zg < self.tiny] = self.tiny
                r_diff_zg2 = r_diff_zg ** 2
                u_g = self.r12_a[water_oxygen_ind[0]] / r_diff_zg2 ** 6
                self.u_g += u_g.copy()
                u_g /= r_diff_zg2
                r_diff_zg *= u_g.copy()
                self.grad_u_vg[2, :, :, :] += r_diff_zg

            else:
                # Ghost atoms are added below the explicit water layer
                cell = atoms.cell.copy() / Bohr
                cell[2][2] = 1.
                natoms_in_plane = [round(np.linalg.norm(cell[0]) * 1.5),
                                   round(np.linalg.norm(cell[1]) * 1.5)]

                plane_z = (oxygen[:, 2].min() - 1.75 *
                           self.atomic_radii_output[water_oxygen_ind[0]])
                nghatoms_z = int(round(oxygen[:, 2].min() -
                                 atoms.positions[:, 2].min()))

                for i in range(int(natoms_in_plane[0])):
                    for j in range(int(natoms_in_plane[1])):
                        for k in np.linspace(atoms.positions[:, 2].min(),
                                             plane_z, num=nghatoms_z):

                            O_layer.append(np.dot(np.array(
                                [(1.5 * i - natoms_in_plane[0] / 4) /
                                 natoms_in_plane[0],
                                 (1.5 * j - natoms_in_plane[1] / 4) /
                                 natoms_in_plane[1],
                                 k / Bohr]), cell))

                # Add additional ghost O-atoms below the actual water O atoms
                # of water which frees the interface in case of corrugated
                # water layers
                for ox in oxygen / Bohr:
                    O_layer.append([ox[0], ox[1], ox[2] - 1.0 *
                                    self.atomic_radii_output[
                                        water_oxygen_ind[0]] / Bohr])

                r12_add = []
                for i in range(len(O_layer)):
                    self.pos_aav[len(atoms) + i] = [O_layer[i]]
                    r12_add.append(self.r12_a[water_oxygen_ind[0]])
                r12_add = np.array(r12_add)
                # r12_a must have same dimensions as pos_aav items
                self.r12_a = np.concatenate((self.r12_a, r12_add))

        for index, pos_av in self.pos_aav.items():
            pos_av = np.array(pos_av)
            r12 = self.r12_a[index]
            for pos_v in pos_av:
                origin_vg = pos_v[:, na, na, na]
                r_diff_vg = self.r_vg - origin_vg
                r_diff2_g = (r_diff_vg ** 2).sum(0)
                r_diff2_g[r_diff2_g < self.tiny] = self.tiny
                u_g = r12 / r_diff2_g ** 6
                self.u_g += u_g
                u_g /= r_diff2_g
                r_diff_vg *= u_g[na, ...]
                self.grad_u_vg += r_diff_vg

        self.u_g *= self.u0 / Ha
        self.grad_u_vg *= -12. * self.u0 / Ha
        self.grad_u_vg[self.grad_u_vg < -1e20] = -1e20
        self.grad_u_vg[self.grad_u_vg > 1e20] = 1e20

        return True


class SJM_RealSpaceHamiltonian(SolvationRealSpaceHamiltonian):
    """Realspace Hamiltonian with continuum solvent model in the context of
    SJM.

    See also Section III of
    A. Held and M. Walter, J. Chem. Phys. 141, 174108 (2014).

    In contrast to the standard implicit solvent model a dipole correction can
    also be applied; this is the only difference from its parent.
    """

    def __init__(self, cavity, dielectric, interactions, gd, finegd, nspins,
                 setups, timer, xc, world, redistributor, vext=None,
                 psolver=None, stencil=3, collinear=None):

        self.cavity = cavity
        self.dielectric = dielectric
        self.interactions = interactions
        cavity.set_grid_descriptor(finegd)
        dielectric.set_grid_descriptor(finegd)
        for ia in interactions:
            ia.set_grid_descriptor(finegd)

        if psolver is None:
            psolver = WeightedFDPoissonSolver()
            self.dipcorr = False
        elif isinstance(psolver, dict):
            psolver = SJMDipoleCorrection(WeightedFDPoissonSolver(),
                                          psolver['dipolelayer'])
            self.dipcorr = True

        if self.dipcorr:
            psolver.poissonsolver.set_dielectric(self.dielectric)
        else:
            psolver.set_dielectric(self.dielectric)

        self.gradient = None

        RealSpaceHamiltonian.__init__(
            self,
            gd, finegd, nspins, collinear, setups, timer, xc, world,
            vext=vext, psolver=psolver,
            stencil=stencil, redistributor=redistributor)

        for ia in interactions:
            setattr(self, 'e_' + ia.subscript, None)
        self.new_atoms = None
        self.vt_ia_g = None
        self.e_total_free = None
        self.e_total_extrapolated = None

    def initialize(self):
        if self.dipcorr:
            self.gradient = [Gradient(self.finegd, i, 1.0,
                             self.poisson.poissonsolver.nn)
                             for i in (0, 1, 2)]
        else:
            self.gradient = [Gradient(self.finegd, i, 1.0,
                             self.poisson.nn)
                             for i in (0, 1, 2)]

        self.vt_ia_g = self.finegd.zeros()
        self.cavity.allocate()
        self.dielectric.allocate()
        for ia in self.interactions:
            ia.allocate()
        RealSpaceHamiltonian.initialize(self)


class CavityShapedJellium(Jellium):
    """The jellium object, where the counter charge takes the form of the
    solvent cavity. It puts the jellium background charge where the solvent is
    present and z < z2.

    Parameters:
    ----------
    charge: float
        The total jellium background charge.
    g_g: array
        The g function from the implicit solvent model, representing the
        percentage of the actual dielectric constant on the grid.
    z2: float
        Position of upper surface in Angstrom units.
    """

    def __init__(self, charge, g_g, z2):

        Jellium.__init__(self, charge)
        self.g_g = g_g
        self.z2 = (z2 - 0.0001) / Bohr

    def todict(self):
        dct = Jellium.todict(self)
        dct.update(z2=self.z2 * Bohr + 0.0001,
                   z1='cavity_like')
        return dct

    def get_mask(self):
        r_gv = self.gd.get_grid_point_coordinates().transpose((1, 2, 3, 0))
        mask = np.logical_not(r_gv[:, :, :, 2] > self.z2).astype(float)
        mask *= self.g_g
        return mask


class SJMDipoleCorrection(DipoleCorrection):
    """Dipole-correcting wrapper around another PoissonSolver specific for SJM.

    Iterative dipole correction class as applied in SJM.

    Notes
    -----

    The modules can easily be incorporated in the trunk version of GPAW
    by just adding the `fd_solv_solve`  and adapting the `solve` modules
    in the `DipoleCorrection` class.

    This module is currently calculating the correcting dipole potential
    iteratively and we would be very grateful if anybody could
    provide an analytical solution.

    New Parameters
    ---------
    corrterm: float
        Correction factor for the added countering dipole. This is calculated
        iteratively.

    last_corrterm: float
        Corrterm in the last iteration for getting the change of slope with
        change in corrterm

    last_slope: float
        Same as for `last_corrterm`

    """
    def __init__(self, poissonsolver, direction, width=1.0):
        """Construct dipole correction object."""

        DipoleCorrection.__init__(self, poissonsolver, direction, width=1.0)
        self.corrterm = 1
        self.elcorr = None
        self.last_corrterm = None

    def solve(self, pot, dens, **kwargs):
        if isinstance(dens, np.ndarray):
            # finite-diference Poisson solver:
            if hasattr(self.poissonsolver, 'dielectric'):
                return self.fd_solv_solve(pot, dens, **kwargs)
            else:
                return self.fdsolve(pot, dens, **kwargs)
        # Plane-wave solver:
        self.pwsolve(pot, dens)

    def fd_solv_solve(self, vHt_g, rhot_g, **kwargs):

        gd = self.poissonsolver.gd
        slope_lim = 1e-8
        slope = slope_lim * 10

        dipmom = gd.calculate_dipole_moment(rhot_g)[2]

        if self.elcorr is not None:
            vHt_g[:, :] -= self.elcorr

        iters2 = self.poissonsolver.solve(vHt_g, rhot_g, **kwargs)

        sawtooth_z = self.sjm_sawtooth()
        L = gd.cell_cv[2, 2]

        while abs(slope) > slope_lim:
            vHt_g2 = vHt_g.copy()
            self.correction = 2 * np.pi * dipmom * L / \
                gd.volume * self.corrterm
            elcorr = -2 * self.correction

            elcorr *= sawtooth_z
            elcorr2 = elcorr[gd.beg_c[2]:gd.end_c[2]]
            vHt_g2[:, :] += elcorr2

            VHt_g = gd.collect(vHt_g2, broadcast=True)
            VHt_z = VHt_g.mean(0).mean(0)
            slope = VHt_z[2] - VHt_z[10]

            if abs(slope) > slope_lim:
                if self.last_corrterm is not None:
                    ds = (slope - self.last_slope) / \
                        (self.corrterm - self.last_corrterm)
                    con = slope - (ds * self.corrterm)
                    self.last_corrterm = self.corrterm
                    self.corrterm = -con / ds
                else:
                    self.last_corrterm = self.corrterm
                    self.corrterm -= slope * 10.
                self.last_slope = slope
            else:
                vHt_g[:, :] += elcorr2
                self.elcorr = elcorr2

        return iters2

    def sjm_sawtooth(self):
        gd = self.poissonsolver.gd
        c = self.c
        L = gd.cell_cv[c, c]
        step = gd.h_cv[c, c] / L

        eps_g = gd.collect(self.poissonsolver.dielectric.eps_gradeps[0],
                           broadcast=True)
        eps_z = eps_g.mean(0).mean(0)

        saw = [-0.5]
        for i, eps in enumerate(eps_z):
            saw.append(saw[i] + step / eps)
        saw = np.array(saw)
        saw /= saw[-1] + step / eps_z[-1] - saw[0]
        saw -= (saw[0] + saw[-1] + step / eps_z[-1]) / 2.
        return saw


class PotentialConvergenceError(ConvergenceError):
    """Raised if potential did not equilibrate."""



import numpy as np

from gpaw.response import ResponseContext
from gpaw.response.pair_integrator import PairFunctionIntegrator
from gpaw.response.pair_functions import PairFunction
from gpaw.response.chiks import get_temporal_part
from gpaw.response.frequencies import ComplexFrequencyDescriptor


class JDOS(PairFunction):

    def __init__(self, spincomponent, qpd, zd):
        self.spincomponent = spincomponent
        super().__init__(qpd, zd)
    
    def zeros(self):
        nz = len(self.zd)
        return np.zeros(nz, float)


class JDOSCalculator(PairFunctionIntegrator):
    r"""Joint density of states calculator.

    Here, the joint density of states of collinear systems is defined as the
    spectral part of the four-component Kohn-Sham susceptibility,
    see [PRB 103, 245110 (2021)]:

                   __  __
                1  \   \   /
    g^μν(q,ω) = ‾  /   /   | σ^μ_ss' σ^ν_s's (f_nks - f_n'k+qs')
                V  ‾‾  ‾‾  \
                   k   t                                \
                             x δ(ħω - [ε_n'k's'-ε_nks]) |
                                                        /

    where t is a composite band and spin transition index: (n, s) -> (n', s').
    """

    def __init__(self, gs, context=None,
                 nbands=None, bandsummation='pairwise',
                 **kwargs):
        """Contruct the JDOSCalculator

        Parameters
        ----------
        gs : ResponseGroundStateAdapter
        context : ResponseContext
        nbands : int
            Number of bands to include in the sum over states
        bandsummation : str
            Band summation strategy (does not change the result, but can affect
            the run-time).
            'pairwise': sum over pairs of bands
            'double': double sum over band indices.
        kwargs : see gpaw.response.pair_integrator.PairFunctionIntegrator
        """
        if context is None:
            context = ResponseContext()
        assert isinstance(context, ResponseContext)

        super().__init__(gs, context, **kwargs)

        self.nbands = nbands
        self.bandsummation = bandsummation

    def calculate(self, spincomponent, q_c, zd):
        """Calculate g^μν(q,ω) using a lorentzian broadening of the δ-function

        Parameters
        ----------
        spincomponent : str
            Spin component (μν) of the joint density of states.
            Currently, '00', 'uu', 'dd', '+-' and '-+' are implemented.
        q_c : list or np.array
            Wave vector in relative coordinates
        zd : ComplexFrequencyDescriptor
            Complex frequencies ħz to evaluate g^μν(q,ω) at, where z = ω + iη
            and η > 0 yields the HWHM of the resulting lorentzian broadening.
        """
        assert isinstance(zd, ComplexFrequencyDescriptor)
        assert zd.upper_half_plane

        # Prepare to sum over bands and spins
        transitions = self.get_band_and_spin_transitions(
            spincomponent, nbands=self.nbands,
            bandsummation=self.bandsummation)

        self.context.print(self.get_info_string(
            q_c, len(zd), spincomponent, self.nbands, len(transitions)))

        # Set up output data structure
        # We need a dummy plane-wave descriptor (without plane-waves, hence the
        # vanishing ecut) for the PairFunctionIntegrator to be able to analyze
        # the symmetries of the system and reduce the k-point integration
        qpd = self.get_pw_descriptor(q_c, ecut=1e-3)
        jdos = JDOS(spincomponent, qpd, zd)

        # Perform actual in-place integration
        self.context.print('Integrating the joint density of states:')
        self._integrate(jdos, transitions)

        return jdos

    def add_integrand(self, kptpair, weight, jdos):
        r"""Add the g^μν(q,ω) integrand of the outer k-point integral:
                        __
                  -1    \  σ^μ_ss' σ^ν_s's (f_nks - f_n'k's')
        (...)_k = ‾‾ Im /  ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
                  π     ‾‾   ħω - (ε_n'k's' - ε_nks) + iħη
                        t

        NB: Since the implemented spin matrices are real, the dissipative part
        is equal to the imaginary part (up to a factor of π) of the full
        integrand.
        """
        # Specify notation
        jdos_z = jdos.array

        # Extract the temporal ingredients from the KohnShamKPointPair
        transitions = kptpair.transitions  # transition indices (n,s)->(n',s')
        df_t = kptpair.df_t  # (f_n'k's' - f_nks)
        deps_t = kptpair.deps_t  # (ε_n'k's' - ε_nks)

        # Construct jdos integrand via the imaginary part of the frequency
        # dependence in χ_KS^μν(q,z)
        if jdos.spincomponent == '00' and self.gs.nspins == 1:
            weight = 2 * weight
        x_zt = get_temporal_part(jdos.spincomponent, jdos.zd.hz_z,
                                 transitions, df_t, deps_t,
                                 self.bandsummation)
        integrand_zt = -x_zt.imag / np.pi

        with self.context.timer('Perform sum over t-transitions'):
            jdos_z += weight * np.sum(integrand_zt, axis=1)

    def get_info_string(self, q_c, nz, spincomponent, nbands, nt):
        """Get information about the joint density of states calculation"""
        s = '\n'

        s += 'Calculating the joint density of states with:\n'
        s += '    q_c: [%f, %f, %f]\n' % (q_c[0], q_c[1], q_c[2])
        s += '    Number of frequency points: %d\n' % nz
        s += '    Spin component: %s\n' % spincomponent
        if nbands is None:
            s += '    Bands included: All\n'
        else:
            s += '    Number of bands included: %d\n' % nbands
        s += 'Resulting in:\n'
        s += '    A total number of band and spin transitions of: %d\n' % nt
        s += '\n'

        s += self.get_basic_info_string()

        return s
        



from math import pi

from gpaw.core import UniformGrid
from gpaw.new import zip
from gpaw.new.pot_calc import PotentialCalculator


class UniformGridPotentialCalculator(PotentialCalculator):
    def __init__(self,
                 wf_grid: UniformGrid,
                 fine_grid: UniformGrid,
                 setups,
                 xc,
                 poisson_solver,
                 nct_aR, nct_R,
                 interpolation_stencil_range=3):
        self.fine_grid = fine_grid
        self.nct_aR = nct_aR

        fracpos_ac = nct_aR.fracpos_ac
        atomdist = nct_aR.atomdist

        self.vbar_ar = setups.create_local_potentials(fine_grid, fracpos_ac,
                                                      atomdist)
        self.ghat_aLr = setups.create_compensation_charges(fine_grid,
                                                           fracpos_ac,
                                                           atomdist)

        self.vbar_r = fine_grid.empty()
        self.vbar_ar.to_uniform_grid(out=self.vbar_r)

        n = interpolation_stencil_range
        self.interpolation_stencil_range = n
        self.interpolate = wf_grid.transformer(fine_grid, n)
        self.restrict = fine_grid.transformer(wf_grid, n)

        super().__init__(xc, poisson_solver, setups, nct_R, fracpos_ac)

    def __str__(self):
        txt = super().__str__()
        degree = self.interpolation_stencil_range * 2 - 1
        name = ['linear', 'cubic', 'quintic', 'heptic'][degree // 2]
        txt += ('interpolation: tri-%s ' % name +
                ' # %d. degree polynomial\n' % degree)
        return txt

    def calculate_charges(self, vHt_r):
        return self.ghat_aLr.integrate(vHt_r)

    def calculate_non_selfconsistent_exc(self, nt_sR, xc):
        nt_sr, _, _ = self._interpolate_density(nt_sR)
        vxct_sr = nt_sr.desc.zeros(nt_sr.dims)
        e_xc = xc.calculate(nt_sr, vxct_sr)
        return e_xc

    def _interpolate_density(self, nt_sR):
        nt_sr = self.interpolate(nt_sR)
        if not nt_sR.desc.pbc_c.all():
            Nt1_s = nt_sR.integrate()
            Nt2_s = nt_sr.integrate()
            for Nt1, Nt2, nt_r in zip(Nt1_s, Nt2_s, nt_sr):
                if Nt2 > 1e-14:
                    nt_r.data *= Nt1 / Nt2
        return nt_sr, None, None

    def calculate_pseudo_potential(self, density, vHt_r):
        nt_sr, _, _ = self._interpolate_density(density.nt_sR)
        grid2 = nt_sr.desc

        vxct_sr = grid2.zeros(nt_sr.dims)
        e_xc = self.xc.calculate(nt_sr, vxct_sr)

        charge_r = grid2.empty()
        charge_r.data[:] = nt_sr.data[:density.ndensities].sum(axis=0)
        e_zero = self.vbar_r.integrate(charge_r)

        ccc_aL = density.calculate_compensation_charge_coefficients()

        # Normalize: (LCAO basis functions may extend outside box)
        comp_charge = (4 * pi)**0.5 * sum(ccc_L[0]
                                          for ccc_L in ccc_aL.values())
        comp_charge = ccc_aL.layout.atomdist.comm.sum(comp_charge)
        pseudo_charge = charge_r.integrate()
        charge_r.data *= -(comp_charge + density.charge) / pseudo_charge

        self.ghat_aLr.add_to(charge_r, ccc_aL)

        if vHt_r is None:
            vHt_r = grid2.zeros()
        self.poisson_solver.solve(vHt_r, charge_r)
        e_coulomb = 0.5 * vHt_r.integrate(charge_r)

        vt_sr = vxct_sr
        vt_sr.data += vHt_r.data + self.vbar_r.data
        vt_sR = self.restrict(vt_sr)
        e_kinetic = 0.0
        for spin, (vt_R, nt_R) in enumerate(zip(vt_sR, density.nt_sR)):
            e_kinetic -= vt_R.integrate(nt_R)
            if spin < density.ndensities:
                e_kinetic += vt_R.integrate(self.nct_R)

        e_external = 0.0

        return {'kinetic': e_kinetic,
                'coulomb': e_coulomb,
                'zero': e_zero,
                'xc': e_xc,
                'external': e_external}, vt_sR, vHt_r

    def _move(self, fracpos_ac, atomdist, ndensities):
        self.ghat_aLr.move(fracpos_ac, atomdist)
        self.vbar_ar.move(fracpos_ac, atomdist)
        self.vbar_ar.to_uniform_grid(out=self.vbar_r)
        self.nct_aR.move(fracpos_ac, atomdist)
        self.nct_aR.to_uniform_grid(out=self.nct_R, scale=1.0 / ndensities)

    def force_contributions(self, state):
        density = state.density
        potential = state.potential
        nt_R = density.nt_sR[0]
        vt_R = potential.vt_sR[0]
        if density.ndensities > 1:
            nt_R = nt_R.desc.empty()
            nt_R.data[:] = density.nt_sR.data[:density.ndensities].sum(axis=0)
            vt_R = vt_R.desc.empty()
            vt_R.data[:] = (
                potential.vt_sR.data[:density.ndensities].sum(axis=0) /
                density.ndensities)

        nt_r = self.interpolate(nt_R)
        if not nt_r.desc.pbc_c.all():
            scale = nt_R.integrate() / nt_r.integrate()
            nt_r.data *= scale

        return (self.ghat_aLr.derivative(state.vHt_x),
                self.nct_aR.derivative(vt_R),
                self.vbar_ar.derivative(nt_r))
